#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator PEP-noncompliant PyTorch type hint** unit tests.

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to **PEP-noncompliant PyTorch type hints** (i.e., :mod:`torch`-specific types).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('torch')
def test_decor_torch() -> None:
    '''
    Test that the :func:`beartype.beartype` decorator correctly decorates
    callables annotated by PyTorch-centric type hints.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from collections.abc import Iterable
    from pytest import raises
    from torch import (
        Tensor,
        ones,
    )

    # ....................{ CALLABLES                      }....................
    @beartype
    def open_thine_eyes_eterne(
        and_sphere_them_round: Iterable[Tensor]) -> Iterable[Tensor]:
        '''
        Arbitrary callable both accepting and returning an iterable of PyTorch
        tensors.

        Surprisingly, this callable exercises an edge case unique to PyTorch
        tensors. Why? Because PyTorch tensors satisfy the standard
        :class:`collections.abc.Collection` protocol *but* override the
        ``__bool__()`` dunder method with non-standard behaviour that
        conditionally raises a :exc:`RuntimeError` when the current PyTorch
        tensor contains two or more items. Since *most* PyTorch tensors contain
        two or more items, effectively all PyTorch tensors raise exceptions when
        evaluated in a boolean context.

        This matters, because :mod:`beartype` previously generated optimally
        efficient type-checking code for :obj:`typing.Iterable` type hints.
        How? :mod:`beartype` detected whether the current pith is a non-empty
        collection by testing that pith in a boolean context. Cue destruction
        when the current pith is a PyTorch tensor.

        See Also
        --------
        https://github.com/beartype/beartype/issues/512
            Issue strongly inspiring this unit test.
        '''

        return and_sphere_them_round


    @beartype
    def a_certain_shape_or_shadow(making_way: Iterable[float]) -> None:
        '''
        Arbitrary callable accepting and returning an iterable of simple scalars
        rather than PyTorch tensors.

        This callable exercises a similar (but ultimately distinct) edge case to
        that of the :func:`.open_thine_eyes_eterne` function defined above.

        See Also
        --------
        https://github.com/beartype/beartype/issues/514
            Issue strongly inspiring this unit test.
        '''

        pass

    # ....................{ LOCALS                         }....................
    # Arbitrary non-trivial PyTorch tensor. In this case, this is a
    # one-dimensional tensor of 3 items that are all the floating-point number
    # "1.0".
    with_wings_or_chariot = ones(3)

    # Arbitrary non-trivial PyTorch tensor. In this case, this is a
    # two-dimensional tensor of standard shape 3x3 whose items are all the
    # floating-point number "1.0".
    upon_all_space = ones(3, 3)

    # ....................{ PASS                           }....................
    # Assert that this callable accepts this two-dimensional tensor -- which,
    # interestingly satisfies the definition of "an iterable of one-dimensional
    # tensors" from a certain mathematical perspective that is boring.
    assert open_thine_eyes_eterne(upon_all_space) is upon_all_space

    # ....................{ FAIL                           }....................
    # Assert that this callable raises the expected exception when passed an
    # iterable of non-tensors.
    with raises(BeartypeCallHintParamViolation):
        open_thine_eyes_eterne(
            "Upon all space: space starr'd, and lorn of light;")

    # Assert that this callable raises the expected exception when passed a
    # one-dimensional iterable of tensors.
    with raises(BeartypeCallHintParamViolation):
        a_certain_shape_or_shadow(with_wings_or_chariot)

    # Assert that this callable raises the expected exception when passed an
    # iterable of non-tensors.
    with raises(BeartypeCallHintParamViolation):
        a_certain_shape_or_shadow(
            'With wings or chariot fierce to repossess')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide Click integration tests.

This submodule functionally tests the :mod:`beartype` package against the
third-party :mod:`click` package.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('click')
def test_click_command() -> None:
    '''
    Integration test validating that the :func:`beartype.beartype` decorator
    integrates cleanly with the third-party :class:`click.command` decorator
    and well-known derivatives of that decorator (e.g., the third-party
    :class:`rich_click.command` decorator).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintReturnViolation
    from beartype._util.module.utilmodimport import import_module_attr_or_none
    from click.testing import CliRunner

    # ....................{ LOCALS                         }....................
    # Tuple of the fully-qualified names of all third-package click-centric
    # packages to be dynamically imported and exercised below.
    CLICK_PACKAGE_NAMES = ('click', 'rich_click',)

    # Click test runner with which to exercise these packages.
    click_runner = CliRunner()

    # ....................{ PACKAGES                       }....................
    # For each of these third-package click-centric packages...
    for click_package_name in CLICK_PACKAGE_NAMES:
        # Third-party @command decorator dynamically imported from this package
        # if this package is importable under the active Python interpreter *OR*
        # "None" otherwise (i.e., if this package is unimportable).
        click_command = import_module_attr_or_none(
            module_name=click_package_name, attr_name='command',)

        # If this package is unimportable, continue to the next such package.
        if click_command is None:
            continue
        # Else, this this decorator is importable from this package.

        # ....................{ CALLABLES                  }....................
        @beartype
        @click_command()
        def click_good() -> int:
            '''
            Arbitrary :func:`beartype.beartype`- and
            :func:`click.command`-decorated callable returning a valid value
            satisfying its return type hint.
            '''

            # Feed that face! Feed it good.
            # print('!UGH!')
            return 0xFEEDFACE


        @beartype
        @click_command()
        def click_bad() -> int:
            '''
            Arbitrary :func:`beartype.beartype`- and
            :func:`click.command`-decorated callable returning an invalid value
            violating its return type hint.
            '''

            # Surely, this cannot be!?
            return "A sea of lustre on the horizon's verge"

        # ....................{ PASS                       }....................
        # "click.testing.Result" object produced by invoking this Click test
        # runner against this "good" Click command.
        click_good_result = click_runner.invoke(cli=click_good)
        # print(f'click result: {repr(click_result)}')
        # print(f'click result: {dir(click_result)}')
        # print(f'click result: {click_result.return_value}')
        # print(f'click result: {click_result.exit_code}')

        # Assert that this "good" Click command succeeded without error.
        assert click_good_result.exit_code == 0

        # ....................{ FAIL                       }....................
        # "click.testing.Result" object produced by invoking this Click test
        # runner against this "bad" Click command.
        click_bad_result = click_runner.invoke(cli=click_bad)
        # print(f'click result: {repr(click_result)}')
        # print(f'click result: {click_result.exc_info}')
        # print(f'click result: {dir(click_result.exc_info)}')

        # Assert that this "bad" Click command raised the expected exception.
        # Note that the "click.testing.Result.exc_info" instance variable is a
        # 3-tuple "(exception_type, exception, exception_traceback)".
        assert bool(click_bad_result.exc_info)
        assert click_bad_result.exc_info[0] is BeartypeCallHintReturnViolation



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`591`-compliant **final type hint reducers** (i.e.,
low-level callables converting higher-level type hints created by subscripting
the :obj:`typing.Final` type hint factory to lower-level type hints more readily
consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep591Exception
from beartype._data.typing.datatypingport import Hint
from beartype._util.hint.pep.utilpepget import get_hint_pep_args

# ....................{ REDUCERS                           }....................
#FIXME: Remove *AFTER* deeply type-checking "Final[...]" type hints. For now,
#shallowly type-checking such hints by reduction to their subscripted arguments
#remains the sanest temporary work-around.
def reduce_hint_pep591(hint: Hint, exception_prefix: str) -> Hint:
    '''
    Reduce the passed :pep:`591`-compliant **final type hint** (i.e.,
    subscription of the :obj:`typing.Final` type hint factory) to a lower-level
    type hint currently supported by :mod:`beartype`.

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Final type hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    Hint
        Lower-level type hint currently supported by :mod:`beartype`.

    Raises
    ------
    BeartypeDecorHintPep591Exception
        If this hint is subscripted by two or more child type hints.
    '''

    # Tuple of zero or more child type hints subscripting this type hint.
    hint_args = get_hint_pep_args(hint)

    # Number of child type hints subscripting this type hint.
    hint_args_len = len(hint_args)

    # If this hint is unsubscripted, reduce this hint to the ignorable type hint
    # "object".
    #
    # Note that PEP 591 bizarrely permits the "typing.Final" type hint factory
    # to remain unsubscripted:
    #     * With no type annotation. Example:
    #           ID: Final = 1
    #       The typechecker should apply its usual type inference mechanisms to
    #       determine the type of ID (here, likely, int). Note that unlike for
    #       generic classes this is not the same as Final[Any].
    #
    # Since runtime type-checkers *NEVER* infer types, this permissiveness
    # substantially reduces the usability of this edge case at runtime.
    # Nevertheless, this is a valid edge case. Technically, we could emit a
    # non-fatal warning to recommend the user explicitly type each unsubscripted
    # "typing.Final" type hint. Pragmatically, doing so would only harass large
    # codebases attempting to migrate to @beartype. Doing nothing is preferable.
    if hint_args_len == 0:
        hint = object
    # If, this hint is subscripted by exactly one child type hint, reduce this
    # hint to that child hint.
    elif hint_args_len == 1:
        hint = hint_args[0]
    # Else, this hint is subscripted by two or more child type hints. In this
    # case, raise an exception.
    #
    # Note that "typing.Final" already prohibits subscription by two or more
    # arguments. Ergo, this should *NEVER* happen: e.g.,
    #     >>> import typing
    #     >>> typing.Final[int, float]
    #     TypeError: typing.Final accepts only single type. Got (<class 'int'>,
    #     <class 'float'>).
    else:
        raise BeartypeDecorHintPep591Exception(
            f'{exception_prefix}PEP 591 type hint {repr(hint)} '
            f'erroneously subscripted by {hint_args_len} child type hints.'
        )

    # Return this reduced hint.
    return hint



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`593`-compliant **type hint violation describers** (i.e.,
functions returning human-readable strings explaining violations of
:pep:`593`-compliant :obj:`typing.Annotated` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeCallHintPepRaiseException
from beartype._check.error.errcause import ViolationCause
from beartype._data.hint.sign.datahintsigns import HintSignAnnotated
from beartype._util.hint.pep.proposal.pep593 import (
    get_hint_pep593_metadata,
    get_hint_pep593_metahint,
)
from beartype._data.code.datacodeindent import CODE_INDENT_1
from beartype._util.text.utiltextrepr import represent_pith

# ....................{ FINDERS                            }....................
def find_cause_pep593_annotated(cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either
    satisfies or violates the :pep:`593`-compliant :mod:`beartype`-specific
    **metahint** (i.e., type hint annotating a standard class with one or more
    :class:`beartype.vale._core._valecore.BeartypeValidator` objects, each
    produced by subscripting the :class:`beartype.vale.Is` class or a subclass
    of that class) of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'
    assert cause.hint_sign is HintSignAnnotated, (
        f'{cause.hint_sign} not "HintSignAnnotated".')

    # Defer heavyweight imports.
    from beartype.vale._core._valecore import BeartypeValidator

    # Type hint annotated by this metahint.
    metahint = get_hint_pep593_metahint(cause.hint)

    # Tuple of zero or more arbitrary objects annotating this metahint.
    hint_validators = get_hint_pep593_metadata(cause.hint)

    # Shallow output cause to be returned, type-checking only whether this pith
    # satisfies this metahint.
    # print(f'[593] Finding {cause} shallow cause...')
    cause_shallow = cause.permute_cause_hint_child_insane(metahint).find_cause()

    # If this pith fails to satisfy this metahint, return this cause as is.
    if cause_shallow.cause_str_or_none is not None:
        return cause_shallow
    # Else, this pith satisfies this metahint.

    # Deep output cause to be returned, permuted from this input cause.
    cause_deep = cause.permute_cause()

    # For each beartype validator annotating this metahint...
    for hint_validator in hint_validators:
        # If this is *NOT* a beartype validator, raise an exception.
        #
        # Note that this object should already be a beartype validator, as the
        # @beartype decorator enforces this constraint at decoration time.
        if not isinstance(hint_validator, BeartypeValidator):
            raise _BeartypeCallHintPepRaiseException(
                f'{cause_deep.exception_prefix}PEP 593 type hint '
                f'{repr(cause_deep.hint)} argument {repr(hint_validator)} '
                f'not beartype validator '
                f'(i.e., "beartype.vale.Is*[...]" object).'
            )
        # Else, this is a beartype validator.
        #
        # If this pith fails to satisfy this validator and is thus the cause of
        # this failure...
        elif not hint_validator.is_valid(cause_deep.pith):
            #FIXME: Unit test this up, please.
            # Human-readable string diagnosing this failure.
            hint_diagnosis = hint_validator.get_diagnosis(
                obj=cause_deep.pith,
                indent_level_outer=CODE_INDENT_1,
                indent_level_inner='',
            )

            # Human-readable string describing this failure.
            cause_deep.cause_str_or_none = (
                f'{represent_pith(cause_deep.pith)} violates validator '
                f'{repr(hint_validator)}:\n'
                f'{hint_diagnosis}'
            )

            # Immediately halt iteration.
            break
        # Else, this pith satisfies this validator. Ergo, this validator is
        # *NOT* the cause of this failure. Silently continue to the next.

    # Return this output cause.
    return cause_deep



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook magic** (i.e., global constants widely leveraged
throughout submodules of the :mod:`beartype.claw` subpackage).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.meta import (
    NAME,
    VERSION,
)

# ....................{ STRINGS                            }....................
BEARTYPE_OPTIMIZATION_MARKER = f'{NAME}{VERSION.replace(".", "v")}'
'''
**Beartype optimization marker** (i.e., placeholder substring suffixing the
``optimization`` parameter passed to the magical hidden
:func:`importlib._bootstrap_external.cache_from_source` function with metadata
unique to the currently installed package name and version of :mod:`beartype`).

This marker uniquifies the filename of bytecode files compiled under beartype
import hooks to the abstract syntax tree (AST) transformation applied by this
version of :mod:`beartype`. Why? Because external callers can trivially enable
and disable that transformation for any module by either calling or not calling
beartype import hooks that accept package name arguments (e.g.,
:func:`beartype.claw.beartype_package`) with the name of a package transitively
containing that module. Compiling a beartyped variant of that module to the same
bytecode file as the non-beartyped variant of that module would erroneously
persist beartyping to that module -- even *after* removing the relevant call to
the :func:`beartype.claw.beartype_package` function! Clearly, that's awful.
Enter @agronholm's phenomenal patch, stage left.

Caveats
-------
**Python requires all optimization markers to be alphanumeric strings.** If this
or *any* other optimization marker contains a non-alphanumeric character, Python
raises a fatal exception resembling:

    ValueError: '-beartype-0.14.2' is not alphanumeric

Ergo, this string globally replaces *all* non-alphanumeric characters that are
otherwise commonly present in the version specifier for this version of
:mod:`beartype` by the arbitrary character ``"v`"" (which is *not* present in
the name of this package and thus suitable as a machine-readable delimiter).
'''

# ....................{ STRINGS ~ names                    }....................
BEARTYPE_DECORATOR_FUNC_NAME = '__beartype__'
'''
Unqualified basename of the beartype decorator as imported into the current
user-defined module being imported and thus transformed by the
:class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass.
'''


BEARTYPE_RAISER_FUNC_NAME = '__die_if_unbearable_beartype__'
'''
Unqualified basename of the beartype exception-raiser as imported into the
current user-defined module being imported and thus transformed by the
:class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass.
'''

# ....................{ STRINGS ~ names ~ claw             }....................
BEARTYPE_CLAW_STATE_OBJ_NAME = '__claw_state_beartype__'
'''
Unqualified basename of the beartype import hook state as imported into the
current user-defined module being imported and thus transformed by the
:class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass.
'''


BEARTYPE_CLAW_STATE_CONF_CACHE_VAR_NAME = 'module_name_to_beartype_conf'
'''
Unqualified basename of the **hooked module beartype configuration cache**
(i.e., dictionary mapping from the fully-qualified name of each previously
imported submodule of each package previously registered in our global package
trie to the beartype configuration configuring type-checking by the
:func:`beartype.beartype` decorator of that submodule) relative to the
beartype import hook state, which contains this cache.
'''

# ....................{ STRINGS ~ names : pep : 695        }....................
BEARTYPE_HINT_PEP695_FORWARDREF_ITER_FUNC_NAME = (
    '__iter_hint_pep695_forwardref_beartype__')
'''
Unqualified basename of the :pep:`695`-compliant **type alias unqualified
relative forward reference iterator** (i.e., generator iteratively creating and
yielding one forward reference proxy for each unqualified relative forward
reference in the passed :pep:`695`-compliant type alias  as imported into the
current user-defined module being imported and thus transformed by the
:class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **extraprocess Decidedly Object-Oriented Runtime-checking (DOOR)
type-checking unit tests** (i.e., unit tests exercising :mod:`beartype.door`
functionality within a Python subprocess forked from the active Python process).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ raisers                    }....................
def test_door_extraprocess_multiprocessing(monkeypatch: 'MonkeyPatch') -> None:
    '''
    Test the :class:`beartype.door.die_if_unbearable` raiser function with
    respect to Python subprocesses forked by the standard :mod:`multiprocessing`
    submodule.

    Parameters
    ----------
    monkeypatch : MonkeyPatch
        :mod:`pytest` fixture allowing various state associated with the active
        Python process to be temporarily changed for the duration of this test.

    See Also
    --------
    https://github.com/python/cpython/issues/44791
        Beartype issue exercised by this unit test.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.py.utilpyinterpreter import (
        get_interpreter_command_words)
    from beartype_test._util.command.pytcmdrun import (
        run_command_forward_output,
        run_command_forward_stderr_return_stdout,
    )
    from beartype_test._util.path.pytpathtest import (
        get_test_unit_data_door_extraprocess_dir)

    # ....................{ ASSERTS                        }....................
    # Temporarily change the current working directory (CWD) to the
    # test-specific root directory containing the data package tested below.
    monkeypatch.chdir(get_test_unit_data_door_extraprocess_dir())

    # Tuple of all shell words with which to run a data submodule in this test
    # suite as a script within a Python subprocess forked from the active Python
    # process via the standard command-line option "-m".
    PYTHON_ARGS = get_interpreter_command_words() + (
        # Fully-qualified name of this data submodule with respect to the root
        # directory containing the package containing this submodule.
        '-m', 'data_door_multiprocessing',
    )

    # Attempt to...
    try:
        # Standard output emitted by running this this module as a script,
        # raising an exception on subprocess failure while forwarding all
        # standard error emitted by this subprocess to the standard error file
        # handle of this parent Python process.
        PYTHON_STDOUT = run_command_forward_stderr_return_stdout(
            command_words=PYTHON_ARGS)

        # Assert this output to be the expected output printed by this module.
        assert PYTHON_STDOUT == 'And faster still, beyond all human speed,'
    # If doing so raised *ANY* exception whatsoever...
    except Exception:
        # Re-run the same module as a script except forward both all standard
        # output *AND* error emitted by this subprocess to the standard output
        # and error file handles of this parent Python process. Doing so
        # significantly improves debuggability in the event of a fatal error.
        run_command_forward_output(command_words=PYTHON_ARGS)

        # Re-raise this exception.
        raise



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **NumPy utilities** (i.e., low-level callables handling the
third-party :mod:`numpy` package).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: The top-level of this module should avoid importing from third-party
# optional libraries, both because those libraries cannot be guaranteed to be
# either installed or importable here *AND* because those imports are likely to
# be computationally expensive, particularly for imports transitively importing
# C extensions (e.g., anything from NumPy or SciPy).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.roar import BeartypeLibraryNumpyException
from beartype._data.typing.datatyping import (
    DictStrToType,
    FrozenSetTypes,
    TypeException,
)
from beartype._util.cache.utilcachecall import callable_cached
from string import digits

# ....................{ GETTERS                            }....................
#FIXME: File an upstream NumPy issue politely requesting they publicize either:
#* An equivalent container listing these types.
#* Documentation officially listing these types.
@callable_cached
def get_numpy_dtype_type_abcs() -> FrozenSetTypes:
    '''
    Frozen set of all **NumPy scalar data type abstract base classes** (i.e.,
    superclasses of all concrete NumPy scalar data types (e.g.,
    :class:`numpy.int64`, :class:`numpy.float32`)).

    This getter is memoized for efficiency. To defer the substantial cost of
    importing from NumPy, the frozen set memoized by this getter is
    intentionally deferred to call time rather than globalized as a constant.

    Caveats
    -------
    **NumPy currently provides no official container listing these classes.**
    Likewise, NumPy documentation provides no official list of these classes.
    Ergo, this getter. This has the dim advantage of working but the profound
    disadvantage of inviting inevitable discrepancies between the
    :mod:`beartype` and :mod:`numpy` codebases. So it goes.
    '''

    # Avoid third-party import dependencies.
    from numpy import (  # pyright: ignore
        character,
        complexfloating,
        flexible,
        floating,
        generic,
        inexact,
        integer,
        number,
        signedinteger,
        unsignedinteger,
    )

    # Create, return, and cache a frozen set listing these ABCs.
    return frozenset((
        character,
        complexfloating,
        flexible,
        floating,
        generic,
        inexact,
        integer,
        number,
        signedinteger,
        unsignedinteger,
    ))


#FIXME: Unit test us up, please.
@callable_cached
def get_numpy_dtype_name_sanitized_to_type_reduced() -> DictStrToType:
    '''
    Dictionary mapping from each **sanitized NumPy data type name** (i.e.,
    string value of the :attr:`numpy.dtype.name` instance variable after being
    translated by the :data:`._TRANSLATION_TABLE_DTYPE_NAME_SANITIZER`
    translation table) to the corresponding **reduced type** (i.e.,
    NumPy-agnostic builtin type if a such type corresponds to this data type
    *or* the direct NumPy-specific abstract base class (ABC) of this data type
    otherwise).

    This getter is memoized for efficiency. To defer the substantial cost of
    importing from NumPy, the frozen set memoized by this getter is
    intentionally deferred to call time rather than globalized as a constant.
    '''

    # Defer heavyweight imports.
    from numpy import unsignedinteger  # pyright: ignore

    # Dictionary mapping from each sanitized NumPy data type name to the
    # corresponding reduced type.
    #
    # Note that the family of "numpy.uint*" data types are the *ONLY* data types
    # lacking a corresponding builtin type, interestingly. Since Python itself
    # has no concept of an "unsigned integer" the NumPy-specific
    # "unsignedinteger" ABC is preferred instead.
    _DTYPE_NAME_SANITIZED_TO_BUILTIN_TYPE = {
        'bool': bool,
        'bytes': bytes,
        'complex': complex,
        'float': float,
        'int': int,
        'uint': unsignedinteger,
        'str': str,
        'void': memoryview,
    }

    # Return this dictionary.
    return _DTYPE_NAME_SANITIZED_TO_BUILTIN_TYPE

# ....................{ REDUCERS                           }....................
#FIXME: Unit test us up, please.
def reduce_numpy_dtype(
    # Mandatory parameters.
    dtype: object,

    # Optional parameters.
    exception_prefix: str = '',
    exception_cls: TypeException = BeartypeLibraryNumpyException,
) -> type:
    '''
    Reduce the passed fine-grained **NumPy data type** (i.e., third-party
    :class:`numpy.dtype` object like :obj:`numpy.complex128`) to a
    coarse-grained **NumPy-agnostic builtin type** (e.g., :class:`complex`) if
    possible *or* simply return this data type as is otherwise (i.e., if this
    data type *cannot* be reduced to a builtin type).

    Parameters
    ----------
    dtype : object
        NumPy data type to be reduced.
    exception_prefix : str, optional
        Human-readable label prefixing raised exception messages. Defaults to
        the empty string.
    exception_cls : Type[Exception], optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeLibraryNumpyException`.

    Returns
    -------
    type
        Either:

        * If this data type is reducible to a NumPy-agnostic builtin type, that
          builtin type.
        * Else, this data type as is.

    Raises
    ------
    exception_cls
        If this NumPy data type is *not* actually a NumPy data type.
    '''

    # Proper dtype coerced from this possibly non-dtype.
    dtype = make_numpy_dtype(
        dtype=dtype,
        exception_prefix=exception_prefix,
        exception_cls=exception_cls,
    )

    # Sanitized name of this dtype, efficiently truncating *ALL* digits and
    # underscores from this name (e.g., sanitizing "complex128" to "complex").
    dtype_name_sanitized = dtype.name.translate(  # type: ignore[attr-defined]
        _TRANSLATION_TABLE_DTYPE_NAME_SANITIZER) 

    # Dictionary mapping from each sanitized NumPy data type name to the
    # corresponding reduced type (e.g., from the string "complex" to the
    # builtin type "complex").
    DTYPE_NAME_SANITIZED_TO_TYPE_REDUCED = (
        get_numpy_dtype_name_sanitized_to_type_reduced())

    # Possibly NumPy-agnostic builtin type reduced from this dtype if this dtype
    # is reducible to such a type *OR* this dtype as is otherwise.
    type_reduced = DTYPE_NAME_SANITIZED_TO_TYPE_REDUCED.get(
        dtype_name_sanitized, dtype)

    # Return this reduced type.
    return type_reduced

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
def make_numpy_dtype(
    # Mandatory parameters.
    dtype: object,

    # Optional parameters.
    exception_prefix: str = '',
    exception_cls: TypeException = BeartypeLibraryNumpyException,
) -> type:
    '''
    **NumPy data type** (i.e., third-party :class:`numpy.dtype` instance)
    coerced from the passed arbitrary object.

    This factory is effectively memoized due to the underlying
    :meth:`numpy.dtype.__new__` constructor already being memoized.

    Parameters
    ----------
    dtype : object
        Object to be coerced into a NumPy data type.
    exception_prefix : str, optional
        Human-readable label prefixing raised exception messages. Defaults to
        the empty string.
    exception_cls : Type[Exception], optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeLibraryNumpyException`.

    Parameters
    ----------
    numpy.dtype
        NumPy data type coerced from this object.

    Raises
    ------
    exception_cls
        If this object is *not* coercible into a NumPy data type.
    '''
    assert isinstance(exception_prefix, str), (
        f'{repr(exception_prefix)} not string.')
    assert isinstance(exception_cls, type), (
        f'{repr(exception_cls)} not exception type.')

    # Defer heavyweight imports.
    from numpy import dtype as numpy_dtype  # pyright: ignore

    # Attempt to coerce this possibly non-dtype into a proper dtype.
    #
    # Note that the dtype.__init__() constructor efficiently maps non-dtype
    # scalar types (e.g., "numpy.float64") to corresponding cached dtypes:
    #     >>> import numpy
    #     >>> i4_dtype = numpy.dtype('>i4')
    #     >>> numpy.dtype(i4_dtype) is numpy.dtype(i4_dtype)
    #     True
    #     >>> numpy.dtype(numpy.float64) is numpy.dtype(numpy.float64)
    #     True
    #
    # Ergo, the call to this constructor here is guaranteed to already
    # effectively be memoized.
    try:
        dtype = numpy_dtype(dtype)  # type: ignore[call-overload]
    # If this object is *NOT* coercible into a dtype, raise an exception. This
    # is essential. As of NumPy 1.21.0, "numpy.typing.NDArray" fails to validate
    # its subscripted argument to actually be a dtype: e.g.,
    #     >>> from numpy.typing import NDArray
    #     >>> NDArray['wut']
    #     numpy.ndarray[typing.Any, numpy.dtype['wut']]  # <-- you kidding me?
    except Exception as exception:
        raise exception_cls(
            f'{exception_prefix}NumPy data type {repr(dtype)} invalid '
            f'(i.e., neither data type nor coercible into data type).'
        ) from exception

    # Return this dtype.
    return dtype  # type: ignore[return-value]

# ....................{ PRIVATE ~ constants                }....................
_TRANSLATION_TABLE_DTYPE_NAME_SANITIZER = str.maketrans('', '', digits + '_')
'''
**Translation table** (i.e., object suitable for passing to the standard
:meth:`str.translate` method as the sole parameter) sanitizing arbitrary **NumPy
data type** (i.e., third-party :class:`numpy.dtype` object) names.

This table strips all ignorable trailing digits and underscores from NumPy data
type names, efficiently reducing the names of NumPy data types (e.g.,
``"complex128"``, ``"str_"``) to the corresponding names of builtin types (e.g.,
``"complex"``, ``"str"``).

See Also
--------
https://stackoverflow.com/a/12856384/2809027
    StackOverflow answer strongly inspiring this implementation.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Python expression indentation substrings** (i.e., string
constants intended to be embedded as syntactically valid indentation in
dynamically generated Python expressions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ PRIVATE ~ subclasses               }....................
class _IndentLevelToCode(dict):
    '''
    **Indentation cache** (i.e., dictionary mapping from 1-based indentation
    levels to the corresponding indentation string constant).

    See Also
    --------
    :data:`.INDENT_LEVEL_TO_CODE`
        Singleton instance of this dictionary subclass.
    '''

    # ....................{ DUNDERS                        }....................
    def __missing__(self, indent_level: int) -> str:
        '''
        Dunder method explicitly called by the superclass
        :meth:`dict.__getitem__` method implicitly called on the first ``[``-
        and ``]``-delimited attempt to access an indentation string constant
        with the passed indentation level.

        Parameters
        ----------
        indent_level : int
            1-based level of indentation to be created, cached, and returned.

        Returns
        -------
        str
            String constant indented to this level of indentation.

        Raises
        ------
        AssertionError
            If either:

            * ``indent_level`` is *not* an integer.
            * ``indent_level`` is a **non-positive integer** (i.e., is less than
              or equal to 0).
        '''
        assert isinstance(indent_level, int), (
            f'{repr(indent_level)} not integer.')
        assert indent_level > 0, f'{indent_level} <= 0.'
        # print(f'Generating indentation level {indent_level}...')

        # String constant indented to this level of indentation.
        #
        # Note that this could also be done recursively (e.g., as
        # "self[indent_level - 1]"), but that doing so would be needlessly cute,
        # overly slow, and dangerously fragile for *NO* good reason.
        indent_code = '    ' * indent_level

        # Cache this string constant.
        self[indent_level] = indent_code

        # Return this string constant.
        return indent_code

# ....................{ MAPPINGS                           }....................
INDENT_LEVEL_TO_CODE = _IndentLevelToCode()
'''
**Indentation cache singleton** (i.e., global dictionary efficiently mapping
from 1-based indentation levels to the corresponding indentation string
constant).

Caveats
-------
**Indentation string constants should always be accessed via this cache rather
than manually generated.** This cache dynamically creates and efficiently caches
indentation string constants on the first access of those constants, obviating
the performance cost of string formatting required to create these constants.

Examples
--------
.. code-block:: pycon

   >>> from beartype._data.code.datacodeindent import INDENT_LEVEL_TO_CODE
   >>> INDENT_LEVEL_TO_CODE[1]
   '    '
   >>> INDENT_LEVEL_TO_CODE[2]
   '        '
'''

# ....................{ STRINGS                            }....................
CODE_INDENT_1 = INDENT_LEVEL_TO_CODE[1]
'''
Code snippet expanding to a single level of indentation.
'''


CODE_INDENT_2 = INDENT_LEVEL_TO_CODE[2]
'''
Code snippet expanding to two levels of indentation.
'''


CODE_INDENT_3 = INDENT_LEVEL_TO_CODE[3]
'''
Code snippet expanding to three levels of indentation.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`649` **type hint utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep649` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_pep649_hintable_annotations() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep649.get_pep649_hintable_annotations`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep649Exception
    from beartype.typing import Optional
    from beartype._data.func.datafuncarg import ARG_NAME_RETURN
    from beartype._util.hint.pep.proposal.pep649 import (
        get_pep649_hintable_annotations)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_14
    from pytest import raises

    # If the active Python interpreter targets Python >= 3.14 and thus supports
    # PEP 649...
    if IS_PYTHON_AT_LEAST_3_14:
        # Defer version-specific imports.
        from beartype_test.a00_unit.data.pep.data_pep649 import (
            unit_test_get_pep649_hintable_annotations)

        # Perform this test.
        unit_test_get_pep649_hintable_annotations()
    # Else, this interpreter targets Python < 3.14 and thus fails to support PEP
    # 649.

    # ....................{ CALLABLES                      }....................
    def too_huge(for_mortal_tongue: set['MoreSorrow']) -> Optional[bytes]:
        '''
        Arbitrary callable accepting one or more parameters and returning one or
        more values -- all annotated by arbitrary type hints *not* containing
        unquoted forward references and thus supported under Python < 3.14.
        '''

        pass

    # ....................{ CLASSES                        }....................
    class MoreSorrow(object):
        '''
        Arbitrary class declaring one or more class variables annotated by
        arbitrary type hints *not* containing unquoted forward references and
        thus supported under Python < 3.14.
        '''

        like_to_this: Optional[str]
        and_such_like_woe: list['MoreSorrow']

    # ....................{ PASS                           }....................
    # Assert this getter passed an annotated callable returns the expected
    # annotations dictionary of that callable.
    assert get_pep649_hintable_annotations(too_huge) == {
        'for_mortal_tongue': set['MoreSorrow'],
        ARG_NAME_RETURN: Optional[bytes],
    }

    # Assert this getter passed an annotated class returns the expected
    # annotations dictionary of that class.
    assert get_pep649_hintable_annotations(MoreSorrow) == dict(
        like_to_this=Optional[str],
        and_such_like_woe=list['MoreSorrow'],
    )

    # ....................{ FAIL                           }....................
    # Assert this getter rejects objects that are *NOT* PEP 649-compliant
    # hintables (i.e., objects that are neither callables, modules, nor types).
    with raises(BeartypeDecorHintPep649Exception):
        get_pep649_hintable_annotations(
            'More sorrow like to this, and such like woe,')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) scope** (i.e., low-level dataclass
aggregating all metadata required to detect and manage a lexical scope being
recursively visited by the current AST transformer).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    # Module,
)
from beartype.claw._ast._scope.clawastscopebefore import (
    BeartypeNodeScopeBeforelist)
from beartype.typing import (
    TYPE_CHECKING,
    Type,
)

# ....................{ CLASSES                            }....................
#FIXME: Unit test us up, please. *sigh*
class BeartypeNodeScope(object):
    '''
    Beartype **abstract syntax tree (AST) scope** (i.e., low-level dataclass
    aggregating all metadata required to detect and manage a lexical scope
    being recursively visited by the current AST transformer).

    Attributes
    ----------
    beforelist : BeartypeNodeScopeBeforelist
        **Abstract syntax tree (AST) scope beforelist** (i.e., dataclass
        aggregating all metadata required to manage the beforelist automating
        decorator positioning for this scope being recursively visited by the
        parent AST transformer).
    name : str
        Fully-qualified name of the current lexical scope (i.e., ``.``-delimited
        absolute name of the module containing this scope followed by the
        relative basenames of zero or more classes and/or callables). This name
        is guaranteed to be prefixed by the current value of the
        :attr:`.module_name` instance variable.
    node_type : type[AST]
        **Lexical scope node type** (i.e., type of the "closest" parent lexical
        scope node of the current node being recursively visited by this AST
        transformer such that that parent node declares a new lexical scope).
        Specifically, if this is:

        * :class:`ast.Module`, the current node directly resides in the **global
          scope** of the current module.
        * :class:`ast.ClassDef`, the current node directly resides in the
          **class scope** of the current class.
        * :class:`ast.FunctionDef`, the current node directly resides in the
          **callable scope** of the current callable.
    _is_beforelist_mutable : bool
        :data:`True` only if this scope's beforelist is **modifiable** (i.e., if
        this scope's beforelist is unique to this scope). Defaults to
        :data:`False` until the caller explicitly calls the
        :meth:`permute_beforelist_if_needed` method, at which point this boolean
        permanently flips to :data:`True` for the lifetime of this scope.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Slot all instance variables defined on this object to reduce the costs of
    # both reading and writing these variables by approximately ~10%.
    __slots__ = (
        'beforelist',
        'name',
        'node_type',
        '_is_beforelist_mutable',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        beforelist: BeartypeNodeScopeBeforelist
        name: str
        node_type: Type[AST]
        _is_beforelist_mutable: bool

    # ....................{ INITIALIZERS                   }....................
    def __init__(
        self,
        beforelist: BeartypeNodeScopeBeforelist,
        name: str,
        node_type: Type[AST],
    ) -> None:
        '''
        Initialize this beforelist scope.

        Parameters
        ----------
        beforelist : BeartypeNodeScopeBeforelist
            **Lexical scope beforelist.** See the class docstring.
        name : str
            Fully-qualified name of the current lexical scope (i.e.,
            ``.``-delimited absolute name of the module containing this scope
            followed by the relative basenames of zero or more classes and/or
            callables).
        node_type : type[AST]
            **Lexical scope node type.** See the class docstring.
        '''
        assert isinstance(beforelist, BeartypeNodeScopeBeforelist), (
            f'{repr(beforelist)} not beforelist scope.')
        assert isinstance(name, str), f'{repr(name)} not string.'
        assert isinstance(node_type, type), f'{repr(node_type)} not type.'
        assert issubclass(node_type, AST), f'{repr(node_type)} not node type.'

        # Classify all passed parameters.
        self.beforelist = beforelist
        self.name = name
        self.node_type = node_type

        # Nullify all remaining instance variables.
        self._is_beforelist_mutable = False

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:

        return '\n'.join((
            f'{self.__class__.__name__}(\n',
            f'    beforelist={repr(self.beforelist)},\n',
            f'    name={repr(self.name)},\n',
            f'    node_type={repr(self.node_type)},\n',
            f'    _is_beforelist_mutable={repr(self._is_beforelist_mutable)},\n',
            f')',
        ))

    # ..................{ BEFORELIST                         }..................
    def permute_beforelist_if_needed(self) -> None:
        '''
        Render this scope's beforelist safe for modification by external callers
        (e.g., to track problematic third-party imports) if this beforelist is
        *not* yet safely **modifiable** (i.e., if this beforelist is still a
        reference to a parent scope's beforelist and is thus *not* unique to
        this scope).

        For both space and time efficiency, beforelists obey the copy-on-write
        design pattern inspired by modern filesystems (e.g., Btrfs). Before
        attempting to modify the contents of this beforelist, callers should
        call this method to render this beforelist safe for modification.
        '''
        # print(f'Permuting scope "{self.name}" if needed...')

        # If this beforelist is *NOT* yet safely modifiable, this beforelist is
        # still a reference to a parent scope's beforelist and is thus *NOT*
        # unique to this scope. In this case...
        if not self._is_beforelist_mutable:
            # print(f'Permuting scope "{self.name}"...')

            # Replace this shared beforelist with a new beforelist unique to
            # this scope, which may then be safely modified by callers.
            self.beforelist = self.beforelist.permute()

            # Record that this beforelist is now safely modifiable.
            self._is_beforelist_mutable = True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`563` + :pep:`604` **integration data submodule.**

This submodule exercises edge cases when combining :pep:`563` via the standard
``from __future__ import annotations`` pragma with :pep:`604`-compliant unions
known to interact problematically with :pep:`563`.
'''

# ....................{ IMPORTS                            }....................
from __future__ import annotations
from beartype import beartype
from dataclasses import dataclass

# ....................{ CLASSES                            }....................
@beartype
@dataclass()
class SpiritMoreVast(object):
    '''
    Arbitrary :pep:`557`-compliant dataclass subclass that exists solely to be
    referred to below in a :pep:`604`-compliant union.
    '''

    than_thine: str = 'Spirit more vast than thine, frame more attuned'
    '''
    Arbitrary field.
    '''


@beartype
@dataclass()
class FrameMoreAttuned(object):
    '''
    Arbitrary :pep:`557`-compliant dataclass subclass that exists solely to be
    referred to below in a :pep:`604`-compliant union.
    '''

    to_beauty: str = 'To beauty, wasting these surpassing powers'
    '''
    Arbitrary field.
    '''


# from beartype import BeartypeConf
# @beartype(conf=BeartypeConf(is_debug=True))
@beartype
@dataclass()
class WithVoiceFarSweeter(object):
    '''
    :pep:`557`-compliant dataclass subclass defining one or more class variables
    annotated by type hints exercising edge cases in :pep:`563` integration
    implemented by the :func:`beartype.beartype` decorator.
    '''

    # ....................{ FIELDS                         }....................
    than_thy_dying_notes: SpiritMoreVastOrFrameMoreAttuned | None  # = None
    '''
    Arbitrary field annotated by an undefined :pep:`604`-compliant union,
    exercising a well-known edge case in :pep:`563` integration.
    '''

# ....................{ HINTS                              }....................
SpiritMoreVastOrFrameMoreAttuned = SpiritMoreVast | FrameMoreAttuned
'''
Arbitrary :pep:`604`-compliant union referenced by the dataclass defined above.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484` **type-checking expression snippets** (i.e., triple-quoted
pure-Python string constants formatted and concatenated together to dynamically
generate boolean expressions type-checking arbitrary objects against
:pep:`484`-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.code.datacodename import (
    ARG_NAME_FUNC,
    VAR_NAME_PITH_ROOT,
)

# ....................{ CODE                               }....................
#FIXME: *FALSE.* The following comment is entirely wrong, sadly. Although that
#comment does, in fact, apply to asynchronous generators, that comment does
#*NOT* apply to coroutines. PEP 484 stipulates that the returns of coroutines
#are annotated in the exact same standard way as the returns of synchronous
#callables are annotated: e.g.,
#   # This is valid, but @beartype currently fails to support this.
#   async def muh_coroutine() -> typing.NoReturn:
#       await asyncio.sleep(0)
#       raise ValueError('Dude, who stole my standards compliance?')
#
#Generalize this snippet to contain a "{{func_call_prefix}}" substring prefixing
#the "{ARG_NAME_FUNC}(*args, **kwargs)" call, please.

# Unlike above, this snippet intentionally omits the "{{func_call_prefix}}"
# substring prefixing the "{ARG_NAME_FUNC}(*args, **kwargs)" call. Why? Because
# callables whose returns are annotated by "typing.NoReturn" *MUST* necessarily
# be synchronous (rather than asynchronous) and thus require no such prefix.
# Why? Because the returns of asynchronous callables are either unannotated
# *OR* annotated by either "Coroutine[...]" *OR* "AsyncGenerator[...]" type
# hints. Since "typing.NoReturn" is neither, "typing.NoReturn" *CANNOT*
# annotate the returns of asynchronous callables. The implication then follows.
PEP484_CODE_CHECK_NORETURN = f'''
    # Call this function with all passed parameters and localize the value
    # returned from this call.
    {VAR_NAME_PITH_ROOT} = {{func_call_prefix}}{ARG_NAME_FUNC}(*args, **kwargs)

    # Since this function annotated by "typing.NoReturn" successfully returned a
    # value rather than raising an exception or halting the active Python
    # interpreter, unconditionally raise an exception.
    #
    # Noop required to artificially increase indentation level. Note that
    # CPython implicitly optimizes this conditional away. Isn't that nice?
    if True'''
'''
:pep:`484`-compliant code snippet calling the decorated callable annotated by
the :attr:`typing.NoReturn` singleton and raising an exception if this call
successfully returned a value rather than raising an exception or halting the
active Python interpreter.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.


'''
:mod:`pytest` **Sphinx test plugin** (i.e., early-time configuration guaranteed
to be run by :mod:`pytest` *after* passed command-line arguments are parsed).

:mod:`pytest` implicitly imports *all* functionality defined by this module
into *all* submodules of this subpackage.
'''

# ....................{ IMPORTS                            }....................
# Attempt to...
try:
    # Import the Sphinx-specific make_app() fixture required to portably test
    # Sphinx documentation builds.
    #
    # Note that the Sphinx-specific test_params() fixture is imported *ONLY* to
    # expose that fixture to make_app(), which requires that fixture.
    from sphinx.testing.fixtures import (
        make_app,
        test_params,
    )
# If doing so fails, silently ignore this failure.
except Exception:
    pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **abstract syntax tree (AST) factory utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.ast.utilastmake` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def node_sibling() -> 'ast.AST':
    '''
    Session-scoped fixture yielding an arbitrary **abstract syntax tree (AST)
    sibling node** (i.e., node intended to be passed as the ``node_sibling``
    keyword parameter to various AST-centric callables defined throughout the
    :mod:`beartype` codebase).
    '''

    # Defer fixture-specific imports.
    from ast import Constant

    # Yield an arbitrary node sibling, defined purely for simplicity as a string
    # literal node whose string is an arbitrary Python identifier.
    yield Constant('In_folds_of_the_green_serpent')

# ....................{ TESTS                              }....................
def test_make_node_object_attr_load(node_sibling: 'ast.AST') -> None:
    '''
    Test the :func:`beartype._util.ast.utilastmake.make_node_object_attr_load`
    factory.

    Parameters
    ----------
    node_sibling : 'ast.AST'
        Arbitrary sibling node to be passed as the ``node_sibling`` parameter.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeClawImportAstException
    from beartype._util.ast.utilastmake import make_node_object_attr_load
    from pytest import raises

    # Assert that this factory raises the expected exception when passed neither
    # the "node_obj" nor "obj_name" parameters.
    with raises(BeartypeClawImportAstException):
        make_node_object_attr_load(
            attr_name='Burn_with_the_poison',
            node_sibling=node_sibling,
        )

    # Assert that this factory raises the expected exception when passed both
    # the "node_obj" and "obj_name" parameters.
    with raises(BeartypeClawImportAstException):
        make_node_object_attr_load(
            attr_name='Burn_with_the_poison',
            node_sibling=node_sibling,
            node_obj=node_sibling,
            obj_name='and_precipitates',
        )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# Note that this is mandatory. If absent, the "autodoc" extension enabled below
# fails with the following build-time error:
#     autodoc: failed to import module 'beartype_sphinx'; the following exception was raised:
#     No module named 'beartype_sphinx'
import os
import sys
sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

project = 'beartype_sphinx'
copyright = '2021, @leycec'
author = '@leycec'

# The full version, including alpha/beta/rc tags
release = '0.0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # ..................{ BUILTIN                            }..................
    # Builtin extensions unconditionally available under *ALL* reasonably
    # modern versions of Sphinx uniquely prefixed by "sphinx.ext.".

    # Builtin extension autogenerating reStructuredText documentation from
    # class, callable, and variable docstrings embedded in Python modules,
    # documenting this project's public (and optionally also private) API.
    'sphinx.ext.autodoc',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# ....................{ TESTS                              }....................
#FIXME: Ugh! This decoration raises non-human-readable exceptions. The culprit
#appears to be PEP 563. For unknown reasons, Sphinx appears to be forcing PEP
#563 for this file in a subversive way. Since this file does *NOT* contain an
#actual "from futures import __annotations__" line, @beartype fails to detect
#that Sphinx has activated PEP 563. The result is pure chaos. We'll probably
#have to do something insane like either:
#* Detect when @beartype is running under Sphinx. We already do that, don't we?
#  Sadly, that detection appears to be failing when run from a "conf.py" file.
#  If that detection actually worked, then this decoration would just silently
#  reduce to a noop -- but it's not, so it's not. *shrug*
#* Detect if the name of the calling script is "conf.py". Ugh! This is terrible.
#
#Clearly, the right thing to do is just reduce @beartype to a noop when run from
#this file -- because the alternative is madness. Ergo, we need to improve our
#Sphinx detection to also detect this file. Shouldn't be terribly arduous... or
#it might be, because Sphinx appears to be exec()-ing this file. (Madness!)
#Still, there should be some deterministic remnant of this perfidious behaviour
#on the call stack, we should think. *sigh*

# from beartype import beartype
#
# @beartype
# def sphinx_conf_beartyped_func() -> str:
#     '''
#     Arbitrary callable annotated by one or more arbitrary type hints.
#
#     This callable exercises an edge case in the :mod:`beartype` codebase in
#     which decorating callables defined in Sphinx-specific ``conf.py`` files
#     previously caused :mod:`beartype` to raise non-human-readable exceptions.
#     '''
#
#     return 'Bother-free is the way to be.'
#
# # Call this function to ensure that it is, indeed, callable.
# sphinx_conf_beartyped_func()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **forward scope classes** (i.e., dictionary subclasses deferring the
resolutions of local and global scopes of classes and callables decorated by the
:func:`beartype.beartype` decorator when dynamically evaluating stringified type
hints for those classes and callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: The BeartypeForwardScope.__init__() "scope_dict: LexicalScope" parameter
#should probably instead be typed as:
#from collections import ChainMap
#...
#    def __init__(self, scope_dict: ChainMap, scope_name: str) -> None:
#
#Why? Because "ChainMap" exists to literally solve this *EXACT* problem.
#Notably, the current approach effectively forces a "BeartypeForwardScope" to
#take a possibly desynchronized "snapshot" of a lexical scope at a certain point
#in time. If either the locals or globals of that scope are subsequently
#modified by an external caller, however, that "BeartypeForwardScope" will then
#be desynchronized from those locals and globals.
#
#A "ChainMap" trivially resolves this. How? Internally, a "ChainMap" only
#holds *REFERENCES* to external locals and globals dictionaries. External
#updates to those external dictionaries are thus *IMMEDIATELY* reflected inside
#the "ChainMap" itself, resolving any desynchronization woes. *facepalm*

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintForwardRefException
from beartype.typing import Type
from beartype._data.typing.datatyping import LexicalScope
from beartype._check.forward.reference.fwdrefabc import (
    _BeartypeForwardRefIndexableABC)
from beartype._check.forward.reference.fwdrefmake import (
    make_forwardref_indexable_subtype)
from beartype._util.func.utilfuncframe import (
    get_frame_caller_module_name_or_none,
    is_frame_caller_beartype,
)
from beartype._util.text.utiltextidentifier import die_unless_identifier

# ....................{ SUBCLASSES                         }....................
#FIXME: Unit test us up, please.
class BeartypeForwardScope(LexicalScope):
    '''
    **Forward scope** (i.e., dictionary mapping from the name to value of each
    locally and globally accessible attribute in the local and global scope of a
    class or callable as well as deferring the resolution of each currently
    undeclared attribute in that scope by replacing that attribute with a
    forward reference proxy resolved only when that attribute is passed as the
    second parameter to an :func:`isinstance`-based runtime type-check).

    This dictionary is principally employed to dynamically evaluate stringified
    type hints, including:

    * :pep:`484`-compliant forward references.
    * :pep:`563`-postponed type hints.

    Attributes
    ----------
    _scope_dict : LexicalScope
        **Composite local and global scope** (i.e., dictionary mapping from
        the name to value of each locally and globally accessible attribute
        in the local and global scope of some class or callable) underlying
        this forward scope. See the :meth:`__init__` method for details.
    _scope_name : str
        Fully-qualified name of this forward scope. See the :meth:`__init__`
        method for details.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        '_scope_dict',
        '_scope_name',
    )

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, scope_dict: LexicalScope, scope_name: str) -> None:
        '''
        Initialize this forward scope.

        Attributes
        ----------
        scope_dict : LexicalScope
            **Composite local and global scope** (i.e., dictionary mapping from
            the name to value of each locally and globally accessible attribute
            in the local and global scope of some class or callable) underlying
            this forward scope.

            Crucially, **this dictionary must composite both the local and
            global scopes for that class or callable.** This dictionary must
            *not* provide only the local or global scope; this dictionary must
            provide both. Why? Because this forward scope is principally
            intended to be passed as the second and last parameter to the
            :func:`eval` builtin, called by the
            :func:`beartype._check.forward.fwdresolve.resolve_hint` function. For
            unknown reasons, :func:`eval` only calls the :meth:`__missing__`
            dunder method of this forward scope when passed only two parameters
            (i.e., when passed only a global scope); :func:`eval` does *not*
            call the :meth:`__missing__` dunder method of this forward scope
            when passed three parameters (i.e., when passed both a global and
            local scope). Presumably, this edge case pertains to the official
            :func:`eval` docstring -- which reads:

                The globals must be a dictionary and locals can be any mapping,
                defaulting to the current globals and locals.
                If only globals is given, locals defaults to it.

            Clearly, :func:`eval` treats globals and locals fundamentally
            differently (probably for efficiency or obscure C implementation
            details). Since :func:`eval` only supports a single unified globals
            dictionary for our use case, the caller *must* composite together
            the global and local scopes into this dictionary. Praise to Guido.
        scope_name : str
            Fully-qualified name of this forward scope. For example:

            * ``"some_package.some_module"`` for a module scope (e.g., to
              resolve a global class or callable against this scope).
            * ``"some_package.some_module.SomeClass"`` for a class scope (e.g.,
              to resolve a nested class or callable against this scope).

        Raises
        ------
        BeartypeDecorHintForwardRefException
            If this scope name is *not* a valid Python attribute name.
        '''
        assert isinstance(scope_dict, dict), (
            f'{repr(scope_dict)} not dictionary.')

        # Initialize our superclass with this lexical scope, efficiently
        # pre-populating this dictionary with all previously declared attributes
        # underlying this forward scope.
        super().__init__(scope_dict)

        # If this scope name is syntactically invalid, raise an exception.
        die_unless_identifier(
            text=scope_name,
            exception_cls=BeartypeDecorHintForwardRefException,
            exception_prefix='Forward scope name ',
        )
        # Else, this scope name is syntactically valid.

        # Classify all passed parameters.
        self._scope_dict = scope_dict
        self._scope_name = scope_name

    # ..................{ DUNDERS                            }..................
    def __missing__(
        self, hint_name: str) -> Type[_BeartypeForwardRefIndexableABC]:
        '''
        Dunder method explicitly called by the superclass
        :meth:`dict.__getitem__` method implicitly called on each ``[``- and
        ``]``-delimited attempt to access an **unresolved type hint** (i.e.,
        *not* currently defined in this scope) with the passed name.

        This dunder method transparently replaces this unresolved type hint with
        a **forward reference proxy** (i.e., concrete subclass of the private
        :class:`beartype._check.forward.reference.fwdrefabc.BeartypeForwardRefABC`
        abstract base class (ABC), which resolves this type hint on the first
        call to the :func:`isinstance` builtin whose second argument is that
        subclass).

        This dunder method assumes that:

        * This scope is only partially initialized.
        * This type hint has yet to be declared in this scope.
        * This type hint will be declared in this scope by the later time that
          this dunder method is called.

        Caveats
        -------
        **This dunder method is susceptible to misuse by third-party frameworks
        that perform call stack inspection.** The higher-level
        :func:`beartype._check.forward.fwdresolve.resolve_hint` internally invokes
        this dunder method by calling the :func:`eval` builtin, which then adds
        a new stack frame to the call stack whose ``f_locals`` and ``f_globals``
        attributes are this dictionary. If a third-party framework introspects
        the call stack containing this new stack frame, this dictionary's
        failure to conform to the behavior of a lexical scope could induce
        failure in that third-party framework. Does this edge case arise in
        real-world usage, though? It does.

        Consider ``pytest``, which detects whether each frame on the call stack
        defines the ``pytest``-specific ``__tracebackhide__`` dunder attribute:

        .. code-block:: python

           def ishidden(self, excinfo: ExceptionInfo[BaseException] | None) -> bool:
               """Return True if the current frame has a var __tracebackhide__
               resolving to True.

               If __tracebackhide__ is a callable, it gets called with the
               ExceptionInfo instance and can decide whether to hide the traceback.

               Mostly for internal use.
               """
               tbh: bool | Callable[[ExceptionInfo[BaseException] | None], bool] = False
               for maybe_ns_dct in (self.frame.f_locals, self.frame.f_globals):
                   # in normal cases, f_locals and f_globals are dictionaries
                   # however via `exec(...)` / `eval(...)` they can be other types
                   # (even incorrect types!).
                   # as such, we suppress all exceptions while accessing __tracebackhide__
                   try:
                       tbh = maybe_ns_dct["__tracebackhide__"]
                   except Exception:
                       pass
                   else:
                       break
               if tbh and callable(tbh):
                   return tbh(excinfo)

        In obscure edge cases involving :mod:`beartype`, ``pytest``, and
        :pep:`563`, one or both of the ``self.frame.f_locals`` and/or
        ``self.frame.f_globals`` dictionaries are instances of the
        :class:`beartype._check.forward.fwdscope.BeartypeForwardScope`
        dictionary subclass. The ``tbh = maybe_ns_dct["__tracebackhide__"]``
        statement then implicitly invokes this dunder method, which then creates
        and returns a new forward reference proxy encapsulating the missing
        ``__tracebackhide__`` attribute: e.g.,

        .. code-block:: python

           tbh = <forwardref__tracebackhide__(
                     __name_beartype__='__tracebackhide__',
                     __scope_name_beartype__='beartype_test.a00_unit.data.pep.pep563.pep695.data_pep563_pep695'
           )>

        Since forward reference proxies are types *and* since types are callable
        (in the sense that "calling" a type instantiates that type), forward
        reference proxies are callable. However, they're not. The
        :class:`beartype._check.forward.reference.fwdrefabc.BeartypeForwardRefABC`
        superclass prohibits instantiation by defining a ``__new__()`` dunder
        method that unconditionally raises exceptions, which then induces the
        ``pytest`` to raise the same exceptions on attempting to
        ``return tbh(excinfo)``.

        This dunder method avoids that and all similar issues by:

        * Detecting whether this dunder method is called by a caller defined
          inside or outside the :mod:`beartype` codebase.
        * If this dunder method is called by a caller defined inside the
          :mod:`beartype` codebase, creating and returning a forward reference
          proxy.
        * If this dunder method is called by a caller defined outside the
          :mod:`beartype` codebase, raising a standard :class:`AttributeError`.

        Parameters
        ----------
        hint_name : str
            Relative (i.e., unqualified) or absolute (i.e., fully-qualified)
            name of this unresolved type hint.

        Returns
        -------
        Type[_BeartypeForwardRefIndexableABC]
            Forward reference proxy deferring the resolution of this unresolved
            type hint.

        Raises
        ------
        BeartypeDecorHintForwardRefException
            If this type hint name is *not* a valid Python attribute name.
        '''
        # print(f'Missing type hint: {repr(hint_name)}')

        # If this type hint name is syntactically invalid, raise an exception.
        die_unless_identifier(
            text=hint_name,
            exception_cls=BeartypeDecorHintForwardRefException,
            exception_prefix='Forward reference ',
        )
        # Else, this type hint name is syntactically valid.

        # If it is *NOT* the case that...
        if not (  # pragma: no cover
            # The caller directly resides inside the "beartype" package *OR*...
            is_frame_caller_beartype(ignore_frames=1) or
            # The caller indirectly resides inside the "beartype" package. This
            # common edge cases arises when the parent
            # beartype._check.forward.fwdresolve.resolve_hint() function calls the
            # eval() builtin to dynamically evaluate the passed stringified type
            # hint: e.g.,
            #     # This is the eval() call triggering this call.
            #     hint_resolved = eval(hint, decor_meta.func_wrappee_scope_forward)
            #
            # In this case, the prior call to the is_frame_caller_beartype()
            # tester tested the stack frame of that eval() call and,
            # specifically, the "__name__" attribute of the global namespace of
            # the external user-defined module proxied by this forward scope.
            # Naturally, that module is external and thus *NOT* inside the
            # "beartype" package. Ignore this stack frame in the hopes that the
            # parent stack frame of that eval() call will be the
            # "beartype._check.forward.fwdresolve" submodule performing that call.
            # Look. We don't like this fragility any more than you do, but
            # Python shenanigans leave us little choice. Our paws are tied!
            is_frame_caller_beartype(ignore_frames=2)
        ):
            # Then the caller is a third-party. In this case, assume this
            # erroneous attempt to access a non-existent attribute of this
            # forward scope to *ACTUALLY* be an Easier to Ask for Permission
            # than Forgiveness (EAFP)-driven to detect whether this forward
            # scope defines this attribute ala the hasattr() builtin. In this
            # case, raise the expected "AttributeError." See the "Caveats"
            # subsection of this dunder method's docstring for commentary.

            # print(f'caller+1: {get_frame_caller_module_name_or_none(ignore_frames=1)}')
            # print(f'caller+2: {get_frame_caller_module_name_or_none(ignore_frames=2)}')
            # print(f'caller+3: {get_frame_caller_module_name_or_none(ignore_frames=3)}')

            # Exception message to be raised.
            exception_message = (
                f'Forward reference scope "{self._scope_name}" '
                f'attribute "{hint_name}" '
            )

            # Fully-qualified name of the module declaring the caller if any
            # *OR* "None" otherwise (e.g., if declared in an interactive REPL).
            frame_caller_module_name = get_frame_caller_module_name_or_none()

            # If the caller has a module, append the fully-qualified name of
            # that module to this exception message to improve debuggability.
            if frame_caller_module_name:
                exception_message += (
                    f'via third-party module "{frame_caller_module_name}" ')
            # Else, the caller has *NO* module.

            # Raise this exception message. Note that we intentionally avoid
            # suffixing the exception message by a "." character here. Why?
            # Because Python treats "AttributeError" exceptions as special.
            # Notably, Python appears to actually:
            # 1. Parse apart the messages of these exceptions for the
            #    double-quoted attribute name embedded in these messages.
            # 2. Suffix these messages by a "." character followed by a sentence
            #    suggesting an existing attribute with a similar name to that of
            #    the attribute name previously parsed from these messages.
            #
            # For example, given an erroneous lookup of a non-existent dunder
            # attribute "__nomnom_beartype__", Python expands the exception
            # message raised below into:
            #     AttributeError: Forward reference scope "MuhRef" dunder
            #     attribute "__nomnom_beartype__" not found. Did you mean:
            #     '__name_beartype__'?
            raise AttributeError(f'{exception_message}not found')
        # Else, the caller resides inside the "beartype" package and is thus
        # assumed to be trustworthy. Don't let us down, @beartype! Not again!

        # Forward reference proxy to be returned.
        forwardref_subtype = make_forwardref_indexable_subtype(
            self._scope_name, hint_name)

        # Cache this proxy.
        self[hint_name] = forwardref_subtype

        # Return this proxy.
        return forwardref_subtype



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import statement abstract syntax tree (AST) transformer mixin**
(i.e., low-level superclass instrumenting import statements relevant to runtime
type-checking in modules hooked by :mod:`beartype.claw` import hooks).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Unit test the newly defined
#"BeartypeDecorPlace.LAST_BEFORE_DECOR_HOSTILE" position, please.

#FIXME: Handle user-defined beforelists via the "self._conf" instance
#variable, please. *sigh*

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    Call,
    ClassDef,
    Import,
    ImportFrom,
    Name,
    alias,
    expr,
    unparse,
)
from beartype.claw._ast._scope.clawastscope import BeartypeNodeScope
from beartype.roar import (
    BeartypeClawAstImportException,
    BeartypeClawImportConfException,
)
from beartype.typing import (
    TYPE_CHECKING,
    Optional,
    Union,
)
from beartype._cave._cavemap import NoneTypeOr
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._conf.decorplace.confplaceenum import BeartypeDecorPlace
from beartype._conf.decorplace.confplacetrie import (
    BeartypeDecorPlaceTrieABC,
    BeartypeDecorPlaceInstanceTrie,
    BeartypeDecorPlaceTypeTrie,
)
from beartype._data.api.standard.dataast import NODE_CONTEXT_LOAD
from beartype._data.conf.dataconfplace import BeartypeDecorPlaceSubtrie
from beartype._data.claw.dataclawmagic import BEARTYPE_DECORATOR_FUNC_NAME
from beartype._data.kind.datakindiota import (
    SENTINEL,
    Iota,
)
# from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._data.typing.datatyping import (
    ListStrs,
    NodeDecoratable,
    NodeVisitResult,
)
from beartype._util.ast.utilastget import (
    get_node_attr_basenames,
    get_node_attr_basename_first,
    get_node_repr_indented,
)
from beartype._util.ast.utilastmunge import copy_node_metadata
# from beartype._util.kind.maplike.utilmapfrozen import FrozenDict
from beartype._util.module.pep.modpep328 import (
    canonicalize_pep328_module_name_relative)

# ....................{ SUBCLASSES                         }....................
class BeartypeNodeTransformerImportMixin(object):
    '''
    Beartype **decorator-hostile tracker** (i.e., visitor pattern recursively
    tracking imports of third-party decorator-hostile decorators as well as the
    modules and types defining such decorators across *all* import statements in
    the AST tree passed to the :meth:`visit` method of the
    :class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass
    also subclassing this mixin).

    This tracker manages the **beartype decorator beforelist** (i.e., collection
    of data structures deciding where the :func:`beartype.beartype` decorator
    should be applied in chains of one or more third-party decorators decorating
    callables and types). This tracker detects third-party decorators well-known
    to be **decorator-hostile** (i.e., decorators hostile to other decorators by
    prematurely terminating decorator chaining such that *no* decorators may
    appear above those decorators in any chain of one or more decorators).

    Attributes
    ----------
    _attr_basenames : Optional[list[str]]
        List of the one or more unqualified basenames comprising the possibly
        fully-qualified name of the current attribute being inspected (e.g.,
        decorator visited by iteration internally performed in the
        :def:`_decorate_node_beartype` method), enabling that inspection to
        reconstruct that name. Specifically, this is either:

        * If the low-level :func:`.get_node_attr_basenames` getter has yet to be
          called by a method of this mixin, :data:`None`.
        * Else, this list.
        '''

    # ..................{ CLASS VARIABLES                    }..................
    # Squelch false negatives from mypy. This is absurd. This is mypy.
    if TYPE_CHECKING:
        _scope: BeartypeNodeScope

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this node transformer mixin.
        '''

        # Initialize our superclass.
        super().__init__()

        # Nullify all instance variables for safety.
        self._attr_basenames: Optional[ListStrs] = None

    # ..................{ MAPPERS                            }..................
    def map_node_attr_imported_to_assigned(
        self, node_name_imported: AST, node_name_assigned: AST) -> None:
        '''
        Map the source attribute (whose possibly fully-qualified name is
        encapsulated by the passed node) previously imported into the current
        lexical scope of the currently visited module to the target attribute
        (whose possibly fully-qualified name is encapsulated by the passed node)
        assigned in this same scope if this source attribute is a third-party
        type transitively defining decorator-hostile decorators *or* silently
        reduce to a noop otherwise (i.e., if this source attribute is *not* such
        a type).

        This public method is intended to be called by external ``visit_*``
        methods of sibling mixins of the
        :class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` class.
        Sibling mixins call this method to notify this decorator-hostile tracker
        of an instantiation of a third-party type known to define
        decorator-hostile decorator methods.

        Parameters
        ----------
        node_name_imported : AST
            Node possibly encapsulating the fully-qualified name or unqualified
            basename of a source type in an assignment statement.
        node_name_assigned : AST
            Node possibly encapsulating the fully-qualified name or unqualified
            basename of a target instance of that type in an assignment
            statement.
        '''
        assert isinstance(node_name_imported, AST), (
            f'{repr(node_name_imported)} not AST node.')
        assert isinstance(node_name_assigned, AST), (
            f'{repr(node_name_assigned)} not AST node.')

        # If this target attribute name is *NOT* encapsulated by a simple "Name"
        # node, this name is *NOT* an unqualified basename. In this case,
        # silently reduce to a noop.
        #
        # Target attribute names that contain one or more "." delimiters are at
        # least partially (and possibly fully) qualified. Handling qualified
        # target attribute names is *EXTREMELY* non-trivial. Doing so
        # necessitates new data structures that have yet to be implemented. See
        # the "FIXME:" comment below for further commentary.
        if not isinstance(node_name_assigned, Name):
            # print(f'Trivially ignoring complex assigned attribute "{unparse(node_name_assigned)}"!')
            return
        # Else, this target attribute name is encapsulated by a simple "Name"
        # node and is thus an unqualified basename.

        # Either:
        # * If this source attribute name refers to a previously imported
        #   decorator-hostile decorator, "True".
        # * If this source attribute name refers to a previously imported
        #   submodule, type, or instance transitively defining one or more
        #   decorator-hostile decorators, the imported decorator-hostile
        #   attribute name subtrie describing the contents of that
        #   submodule, type, or instance.
        # * Else (i.e., if this source attribute name refers to neither a
        #   previously imported decorator-hostile decorator *nor* a
        #   submodule, type, or instance transitively defining these
        #   decorators), "False".
        imported_attr_basename_subtrie = self._is_node_scoped_attr_name(
            node_name_imported)

        # If this source attribute name does *NOT* refer to a previously
        # imported type defining one or more decorator-hostile decorator
        # methods, this assignment statement is *NOT* instantiating an instance
        # of such a type. Ergo, this target attribute *CANNOT* be an instance
        # of such a type. In this case, silently reduce to a noop.
        if not isinstance(
            imported_attr_basename_subtrie, BeartypeDecorPlaceTypeTrie):
            # print(f'Trivially ignoring friendly source attribute "{unparse(node_name_imported)}"!')
            return
        # Else, this source attribute name refers to a previously imported type
        # defining one or more decorator-hostile decorator methods.

        #FIXME: *NICE TRY*... but "Nope." This only works if
        #"node_name_assigned" is a trivial "Name" node. If
        #"node_name_assigned" is instead a non-trivial "Attribute" node
        #hierarchy, we instead need to iterate over this hierarchy and, for each
        #basename comprising a "node_name_assigned_basenames" list of basenames
        #*EXCEPT* the last such basename, instantiate a new
        #"BeartypeDecorPlaceInstanceTrie" wrapping a nested frozen dictionary.
        #
        #But... that's *EXACTLY* the problem, isn't it? Nested frozen dictionary
        #simply don't work. We'd instead need to instantiate mutable nested
        #chain maps so that repeated calls to this method across the same
        #lexical scope behave as expected (i.e., mutate the same nested chain
        #map) rather than raising exceptions (i.e., due to trying to erroneously
        #mutate a nested frozen dictionary).
        #
        #However, that then ignites an even less trivial issue with nested
        #scopes, which then need to somehow be able to efficiently inherit and
        #safely mutate those nested chain maps across nested scopes as needed.
        #In short, the whole thing *REALLY* warrants a whole new
        #"BeartypeScopeChainMap" implementation that pretty much rewrites the
        #standard "ChainMap" type from the ground up to impose recursive chain
        #mapping at *ALL* nesting levels of the chain map hierarchy. By compare,
        #the standard "ChainMap" type only chain maps the outermost nesting
        #level; inner nesting levels are left to their own devices, which is
        #useless for our purposes. We'll probably end up doing that *EVENTUALLY*
        #-- but we really want to delay that eventuality as long as feasible.
        #
        #Consider the common use case of:
        #    from celery import Celery
        #
        #    class SomeClass(object):
        #        def __init__(self) -> None:
        #            # This assignment statement triggers a call to this
        #            # map_node_attr_imported_to_assigned() method!
        #            self._celery = Celery()
        #
        #Interesting. That *IS* the common case, isn't it? "self" attributes. I
        #suppose we could try to support those by manually handling *EXACTLY*
        #the common case of an outer "Attribute" node nesting an inner "Name"
        #node, but... Nah. We'd still need nested chain maps and thus our own
        #"BeartypeScopeChainMap" implementation. Moreover, is there even any
        #point? After all, assigning to "self" semantically assigns to
        #externally instantiated instances of that type. There's *NO* way we're
        #tracking that madness just to detect decorator-hostile decorators,
        #right? Supreme overkill.
        #
        #Ergo, we currently ignore everything *EXCEPT* trivial "Name" nodes.
        assigned_attr_basename_first: str = get_node_attr_basename_first(  # type: ignore[assignment]
            node=node_name_assigned, attr_basenames=self._attr_basenames)

        # Target scoped attribute name subtrie describing all decorator-hostile
        # methods bound to this instance, coerced from this source imported
        # attribute name subtrie for disambiguity. While preserving the type of
        # this subtrie as a "BeartypeDecorPlaceTypeTrie" would (of course) be
        # trivially feasible, doing so could also invite false positives and
        # negatives with subsequent calls to this or other methods incorrectly
        # detecting this instance-specific subtrie as a type-specific subtrie.
        assigned_attr_basename_subtrie = BeartypeDecorPlaceInstanceTrie(
            imported_attr_basename_subtrie)

        # Map the unqualified basename of this target instance of that type in
        # an assignment statement encapsulated by the passed node to the scoped
        # attribute name subtrie describing all decorator-hostile methods bound
        # to this instance.
        self._map_scoped_attr_name_to_subtrie(
            attr_name=assigned_attr_basename_first,
            attr_name_subtrie=assigned_attr_basename_subtrie,
        )

    # ..................{ VISITORS                           }..................
    def visit_Import(self, node: Import) -> NodeVisitResult:
        '''
        Track the passed **import node** (i.e., node signifying the importation
        of a module or package) if this node signifies an import of a module or
        package defining one or more decorator-hostile decorators.

        Parameters
        ----------
        node : Import
            Possibly problematic import node to be tracked.

        Returns
        -------
        NodeVisitResult
            The passed import node unmodified.
        '''

        # ..................{ PREAMBLE                       }..................
        # Recursively transform *ALL* child nodes of this parent node.
        self.generic_visit(node)  # type: ignore[attr-defined]

        # ..................{ SEARCH                         }..................
        # For each child "alias" node of this parent "Import" node encapsulating
        # the fully-qualified names of the one or more modules being imported by
        # this import statement...
        for import_module_alias in node.names:
            # If this child node is *NOT* an "alias", silently skip this
            # unrecognized node type and continue to the next child node.
            #
            # Note that this should *NEVER* happen. All child nodes of parent
            # "Import" nodes should be aliases. For forward compatibility with
            # future Python versions, ignore this unrecognized node type. See
            # similar commentary below for further discussion.
            if not isinstance(import_module_alias, alias):
                continue
            # Else, this child node is an "alias".

            # Fully-qualified name of the external module or package being
            # imported if any *OR* "None".
            #
            # Note that this name should *NEVER* be "None". Nonetheless, mypy
            # claims this name can actually be "None". How, mypy? Even mypy has
            # no idea. To squelch complaints from mypy, we pretend mypy is sane.
            import_module_name = import_module_alias.name
            if not import_module_name: continue  # <-- *SILENCE, MYPY!*

            # Fully-qualified name of the top-level root package or module
            # transitively containing that package or module (e.g.,
            # "some_package" when "import_module_name" is
            # "some_package.some_module.some_submodule").
            #
            # This root package name is *MUCH* more relevant than the exact
            # submodule name (i.e., "import_module_name") imported by this
            # import statement. Why? Because trivial imports of the form "import
            # {package_name}" are similar to complex imports of the form "import
            # {package_name}.{submodule_name}.{attr_name}". In both cases,
            # Python only imports the top-level root package whose name is
            # "{package_name}" into the current lexical scope. For static
            # analysis purposes, trailing "."-prefixed substrings like
            # ".{submodule_name}.{attr_name}" are almost entirely irrelevant.
            # Consequently, all subsequent logic ignores "import_module_name" in
            # favour of this "import_package_name".
            #
            # Note this has been profiled to be the fastest one-liner parsing
            # the first "."-suffixed substring from a "."-delimited string.
            import_package_name = import_module_name.partition('.')[0]

            # If this import statement does *NOT* import a third-party package
            # known to define decorator-hostile decorators, this import
            # statement is ignorable with respect to @beartype. In this case,
            # silently skip this ignorable package by returning this node as is.
            if import_package_name not in (
                self._scope.beforelist.schema_package_names):
                return node
            # Else, this import statement imports a third-party module known to
            # define decorator-hostile decorators.

            # Child (sub)trie of the scoped attribute name trie unique to that
            # package in the current scope of the currently visited module.
            # Note this (sub)trie is guaranteed to exist rather than be "None".
            # Why? Because of the prior validation.
            import_package_name_subtrie = (
                self._scope.beforelist.schema_attr_basename_trie.get(
                    import_package_name))

            # Map the unqualified basename of that package to this subtrie.
            self._map_scoped_attr_name_to_subtrie(
                attr_name=import_package_name,
                attr_name_subtrie=import_package_name_subtrie,
            )

        # ..................{ RETURN                         }..................
        # Return this node unmodified.
        return node


    def visit_ImportFrom(self, node: ImportFrom) -> NodeVisitResult:
        '''
        Track the passed **import-from node** (i.e., node signifying the
        importation of an attribute from a module or package) if this node
        signifies an import of either:

        * A decorator-hostile decorator.
        * A module defining one or more decorator-hostile decorators.
        * A type defining one or more decorator-hostile decorators.

        Parameters
        ----------
        node : ImportFrom
            Possibly problematic import-from node to be tracked.

        Returns
        -------
        NodeVisitResult
            The passed import-from node unmodified.
        '''

        # ..................{ PREAMBLE                       }..................
        # Recursively transform *ALL* child nodes of this parent node.
        self.generic_visit(node)  # type: ignore[attr-defined]

        # ..................{ LOCALS                         }..................
        # Beforelist unique to the currently visited lexical scope of the
        # currently visited module, localized for readability and efficiency.
        beforelist = self._scope.beforelist
        # print(f'beforelist: {beforelist}')

        # Possibly fully-qualified name of the external module being imported
        # from. This name is subsequently canonicalized from both absolute and
        # relative imports (and thus guaranteed to be a non-empty string), but
        # is initially either:
        # * If this is an pure absolute import statement (e.g., "from
        #   some_package.some_module import some_attribute"), the
        #   fully-qualified name of the external module being imported from. In
        #   this case, the "node.level" instance variable is guaranteed to be 0.
        # * If this is a pure relative import statement (e.g., "from .... import
        #   some_attribute"), "None". In this case, the "node.level" instance
        #   variable is guaranteed to be a positive integer.
        # * If this is a mixed absolute-relative import statement (e.g., "from
        #   ....some_submodule import some_attribute"), the partially-qualified
        #   name of the sibling module being imported from. In this case, the
        #   "node.level" instance variable is also a positive integer.
        import_module_name = node.module
        # print(f'\nnode: {unparse(node)}')
        # print(f'import_module_name: {import_module_name}')

        # ..................{ CANONICALIZE                   }..................
        # If this is a PEP 328-compliant relative import statement (e.g., "from
        # .... import some_attribute"), derive the fully-qualified name of the
        # external module being imported from from the fully-qualified name of
        # the currently visited module and the "node.level" instance variable.
        #
        # An example speaks a thousand words. Thus, this code snippet:
        #     from ..... import some_attribute
        #     from .....some_submodule import another_attribute
        #
        # ...has this corresponding abstract syntax tree (AST):
        #     Module(
        #         body=[
        #             ImportFrom(
        #                 names=[
        #                     alias(name='some_attribute')],
        #                 level=5),
        #             ImportFrom(
        #                 module='some_submodule',
        #                 names=[
        #                     alias(name='another_attribute')],
        #                 level=5)])
        if node.level > 0:
            # print(f'module_basenames: {self._module_basenames}')

            # If the top-level root package transitively containing the
            # currently visited module is *NOT* a third-party package known to
            # define decorator-hostile decorators, this import statement
            # relative to that root package is ignorable with respect to
            # @beartype. In this case, silently reduce to a noop by returning
            # this node unmodified.
            #
            # Note that the "_module_basenames" list is guaranteed to be
            # non-empty and thus contain at least a first item, yielding the
            # fully-qualified name of that root package.
            if self._module_basenames[0] not in beforelist.schema_package_names:  # type: ignore[attr-defined]
                # print(f'Ignoring friendly relative import "{unparse(node)}"!')
                return node
            # Else, that root package is known to define decorator-hostile
            # decorators.

            #FIXME: [SPEED] *EXTREMELY INEFFICIENT.* We designed the
            #canonicalize_pep328_module_name_relative() only *AFTER* belatedly
            #realizing that the "ast.ImportFrom" API provided a numeric level
            #rather than a "."-prefixed module name. Instead:
            #* Refactor canonicalize_pep328_module_name_relative() to optionally
            #  accept a new "target_height: Optional[int] = None" parameter.
            #* Probably rename the existing "module_name_relative: str"
            #  parameter to "target_module_name: Optional[str] = None".
            #* Call that function like so below:
            #    import_module_name = canonicalize_pep328_module_name_relative(
            #        source_module_basenames_absolute=self._module_basenames,  # type: ignore[attr-defined]
            #        target_height=node.level,
            #    )
            #
            #Thankfully, this mostly doesn't matter at all. Why? Because
            #*NOBODY* is ever going to both (A) define a package-specific
            #decorator-hostile decorator and (B) import that decorator
            #relatively. Like, really. *ALL* decorator-hostile decorators are
            #defined by external third-party packages and then imported with
            #absolute imports into the current module. This is edge case is
            #beyond rare. Indeed, it's unlikely this will ever trigger in any
            #real-world code anywhere... *EVER*. Oh, well. We sigh so hard!

            # String of one or more "." delimiters corresponding to the package
            # height of this relative import to recurse upward from the
            # fully-qualified name of the currently visited module.
            import_module_name_prefix = '.' * node.level

            # Original "."-prefixed partially-qualified name of the sibling
            # module being imported from (e.g., "....", "....some_submodule"),
            # reconstructed from this "ImportFrom" node, which fails to directly
            # preserve this name. Specifically...
            import_module_name = (
                # If this is a pure relative import, this prefix as is;
                import_module_name_prefix
                if import_module_name is None else
                # Else, this is a mixed absolute-relative import. In this case,
                # this prefix and sibling module name.
                f'{import_module_name_prefix}{import_module_name}'
            )
            # print(f'Parsing unfriendly relative import "{import_module_name}": "{unparse(node)}"...')

            # Canonicalize this imported module name relative to that root
            # package into an absolute imported module name.
            import_module_name = canonicalize_pep328_module_name_relative(
                module_name_relative=import_module_name,
                module_basenames_absolute=self._module_basenames,  # type: ignore[attr-defined]
            )
            # print(f'...into unfriendly absolute import "{import_module_name}"!')
        # Else, this is already an absolute import statement.
        #
        # In either case, the fully-qualified name of the external module being
        # imported from has now been decided.

        # ..................{ LOCALS ~ package               }..................
        # List of each unqualified basename comprising the fully-qualified name
        # of the external module being imported from.
        #
        # Note that:
        # * This list is guaranteed to be non-empty. Why? Because Python syntax
        #   prohibits empty import statements (e.g., "from import ugh\n").
        # * The "str.split('.')" and "str.rsplit('.')" calls produce the same
        #   lists under all edge cases. We arbitrarily call the former rather
        #   than the latter for simplicity.
        import_module_basenames = import_module_name.split('.')  # type: ignore[union-attr]

        # Fully-qualified name of the top-level root package transitively
        # containing that module (e.g., "some_package" when "import_module_name"
        # is "some_package.some_module.some_submodule").
        import_package_name = import_module_basenames[0]

        # If this import statement does *NOT* import from a third-party package
        # known to define decorator-hostile decorators, this import statement is
        # ignorable with respect to @beartype. In this case, silently reduce to
        # a noop by returning this node unmodified.
        if import_package_name not in beforelist.schema_package_names:
            # print(f'Ignoring friendly import "{unparse(node)}"!')
            return node
        # Else, this import statement imports from a third-party package known
        # to define decorator-hostile decorators.
        # print(f'\nParsing unfriendly import "{unparse(node)}"...')

        # ..................{ SEARCH ~ module                }..................
        # Current and prior child (sub)tries of the scoped decorator-hostile
        # attribute name trie of the current scope of the currently visited
        # module iterated over below, efficiently mapping the fully-qualified
        # name of the module being imported from to the corresponding (sub)trie.
        # Specifically, each of these (sub)tries is either:
        # * If the currently iterated basename of that module erroneously refers
        #   to a third-party decorator-hostile decorator callable (which clearly
        #   is *NOT* a method, since callables are *NOT* modules), "None".
        # * Else if that basename refers to a third-party (sub)package or module
        #   transitively defining one or more decorator-hostile decorators, yet
        #   another nested "BeartypeDecorPlaceTrieABC" instance describing the
        #   relevant (sub)package or module structure.
        # * Else, that basename refers to a package or module transitively
        #   defining *NO* decorator-hostile decorators. In this case, the
        #   sentinel placeholder.
        import_module_name_subtrie_parent = (
            beforelist.schema_attr_basename_trie)
        import_module_name_subtrie: (
            Union[BeartypeDecorPlaceSubtrie, Iota]) = None

        # 0-based indices of the current and last unqualified basenames of the
        # fully-qualified name of the external module being imported from.
        import_module_basename_index_curr = 0
        import_module_basename_index_last = len(import_module_basenames) - 1

        # While unqualified basenames of this name remain to be iterated...
        while (
            import_module_basename_index_curr <=
            import_module_basename_index_last
        ):
            # Current unqualified basename of this name.
            import_module_basename = import_module_basenames[
                import_module_basename_index_curr]

            # Either:
            # * If this unqualified basename of the module being imported from
            #   maps to a problematic submodule, type, or instance transitively
            #   defining decorator-hostile decorators, the corresponding subtrie
            #   of this module subtrie.
            # * If this basename maps to a decorator-hostile decorator, "None".
            # * If this basename maps to a friendly attribute (i.e., attribute
            #   that is neither a decorator-hostile decorator *NOR* submodule,
            #   type, or instance transitively defining decorator-hostile
            #   decorators), the sentinel placeholder.
            import_module_name_subtrie = (
                import_module_name_subtrie_parent.get(
                    import_module_basename, SENTINEL))
            # print(f'Visiting imported-from module basename "{import_module_basename}"...')
            # print(f'...associated with subtrie {import_module_name_subtrie}.')

            # If this basename maps to *NO* decorator-hostile module, this
            # import statement imports from *NO* such module and is thus
            # ignorable with respect to @beartype. In this case, silently reduce
            # to a noop by returning this node unmodified.
            if import_module_name_subtrie is SENTINEL:
                # print(f'Ignoring friendly import "{unparse(node)}"!')
                return node
            # Else, this basename maps to a problematic module.
            #
            # If this basename maps to a decorator-hostile decorator, this
            # subtrie erroneously claims that external module being imported
            # from to be a decorator rather than a module! However, Only modules
            # are importable from. Decorators are always either types or
            # callables -- neither of which are modules and thus neither of
            # which are importable from. Clearly, this subtrie (and by
            # extension the scoped decorator-hostile attribute name trie as a
            # whole) is ill-defined. For example, this edge case arises if:
            # * A caller configures the beartype_this_package() hook with a
            #   beartype configuration resembling:
            #     beartype_this_package(BeartypeConf(
            #         schema_attr_basename_trie=FrozenDict({
            #             'bad_package': frozenset(('bad_submodule',)),})))
            # * One or more submodules of the caller's package import attributes
            #   from that submodule resembling:
            #       from bad_package.bad_submodule import bad_attribute
            elif import_module_name_subtrie is None:
                # Pretty-printed representation of this import statement.
                node_repr = get_node_repr_indented(node)

                # Raise an explanatory exception.
                raise BeartypeClawAstImportException(
                    f'Beartype configuration {repr(self._conf)} '  # type: ignore[attr-defined]
                    f'decorator position beforelist '
                    f'{repr(beforelist.schema_attr_basename_trie)} '
                    f'nested frozen dictionary of '
                    f'decorator-hostile decorator names '
                    f'{repr(import_module_name_subtrie_parent)} '
                    f'erroneously implies '
                    f'object "{import_module_name}" to be a '
                    f'decorator rather than a module, but '
                    f'module "{self._module_name}" import statement '  # type: ignore[attr-defined]
                    f'imports from this object as a '
                    f'module rather than a decorator:\n'
                    f'\t{node_repr}'
                )
            # Else, this basename does *NOT* map to a decorator-hostile
            # decorator.
            #
            # Thus, this basename maps to a child subtrie of this parent
            # (sub)trie (by elimination), implying this basename to be that of a
            # third-party module defining decorator-hostile decorators. In this
            # case, attempt to unwrap this subtrie against the next unqualified
            # basename in this hierarchical nesting.

            # Validate sanity.
            assert isinstance(
                import_module_name_subtrie, BeartypeDecorPlaceTrieABC)

            # Preserve this subtrie as the parent subtrie of the next iteration.
            import_module_name_subtrie_parent = import_module_name_subtrie

            # Increment the index of the next unqualified basename to visit.
            import_module_basename_index_curr += 1
        # Note that, by the above logic, the "import_module_name_subtrie" local
        # variable accessed below is now guaranteed to be a non-empty child
        # subtrie (i.e., concrete instance of the "BeartypeDecorPlaceTrieABC"
        # superclass containing one or more key-value pairs) describing a
        # third-party module defining decorator-hostile decorators.

        # ..................{ SEARCH ~ decorator             }..................
        # For each child "alias" node of this parent "ImportFrom" node
        # encapsulating the unqualified basenames of the one or more attributes
        # being imported from that module by this import statement...
        for import_attr_alias in node.names:
            # If this child node is *NOT* an "alias", silently skip this
            # unrecognized node type and continue to the next child node.
            #
            # Note that this should *NEVER* happen. All child nodes of parent
            # "ImportFrom" nodes should be aliases. For forward compatibility
            # with future Python versions, ignore this unrecognized node type.
            # See similar commentary below for further discussion.
            if not isinstance(import_attr_alias, alias):
                # print(f'Ignoring imported-from target non-alias AST node "{unparse(import_attr_alias)}"!')
                continue
            # Else, this child node is an "alias".

            # Imported source attribute name (i.e., unqualified basename of this
            # imported attribute as originally defined inside that external
            # module) if any *OR* "None".
            #
            # Note that this name should *NEVER* be "None". Yet, mypy claims
            # this name can actually be "None". How, mypy? Even mypy has no
            # idea. To squelch mypy complaints, we pretend mypy is still sane.
            import_attr_basename_src = import_attr_alias.name
            if not import_attr_basename_src: continue  # <-- *SILENCE, MYPY!*
            # This name is now guaranteed to be non-"None".

            # Either:
            # * If this unqualified basename of the attribute being imported
            #   from that module maps to a problematic submodule, type, or
            #   instance transitively defining decorator-hostile decorators,
            #   the corresponding subtrie of this module subtrie.
            # * If this basename maps to a decorator-hostile decorator, "None".
            # * If this basename maps to a friendly attribute (i.e., attribute
            #   that is neither a decorator-hostile decorator *NOR* submodule,
            #   type, or instance transitively defining decorator-hostile
            #   decorators), the sentinel placeholder.
            import_attr_name_subtrie = import_module_name_subtrie.get(  # type: ignore[union-attr]
                import_attr_basename_src, SENTINEL)

            # This basename maps to *NO* decorator-hostile attribute, this
            # imported attribute is ignorable with respect to @beartype. In this
            # case, silently continuing to the next imported attribute.
            if import_attr_name_subtrie is SENTINEL:
                # print(f'Ignoring friendly imported-from target attribute "{import_attr_basename_src}"!')
                continue
            # Else, this basename maps to a decorator-hostile attribute.

            # Imported target attribute name (i.e., unqualified basename of this
            # imported attribute as newly localized or globalized inside the
            # currently visited module), defined as either...
            import_attr_basename_trg = (
                # If this import statement imports this attribute under an alias
                # via the 'as' predicate (e.g., "from mcp import tool as
                # mcp_tool"), that alias.
                #
                # Note that the "alias.asname" instance variable is "None" if
                # undefined, according to the official "ast" documentation.
                import_attr_alias.asname or
                # Else, this import statement imports this attribute under its
                # original name (e.g.,  "from mcp import tool"). In this case,
                # that name.
                import_attr_basename_src
            )
            # print(f'Mapping imported decorator-hostile decorator "{import_attr_basename_src}"...')
            # print(f'...aliased to "{import_attr_basename_trg}"...')
            # print(f'...to subtrie "{import_attr_name_subtrie}".')

            # Map the unqualified basename of this decorator-hostile attribute
            # to this attribute's subtrie.
            self._map_scoped_attr_name_to_subtrie(
                attr_name=import_attr_basename_trg,
                attr_name_subtrie=import_attr_name_subtrie,  # type: ignore[arg-type]
            )

        # ..................{ RETURN                         }..................
        # Return this node unmodified.
        return node

    # ....................{ PRIVATE ~ mappers              }....................
    def _map_scoped_attr_name_to_subtrie(
        self,
        attr_name: str,
        attr_name_subtrie: Optional[BeartypeDecorPlaceSubtrie],
    ) -> None:
        '''
        Map the passed possibly fully-qualified name of a third-party
        decorator-hostile attribute accessible to this scope (e.g., by an import
        or assignment statement) to the passed **scoped attribute name subtrie**
        (i.e., recursive tree structure whose nodes are the unqualified
        basenames of third-party attributes imported into a scope of the
        currently visited module such that these attributes are either
        themselves decorator-hostile decorators *or* submodules, types, or
        instances transitively defining decorator-hostile decorators).

        Parameters
        ----------
        attr_name : str
            Possibly fully-qualified name of the attribute to be mapped.
        attr_name_subtrie : Optional[BeartypeDecorPlaceSubtrie]
            Either:

            * If this attribute is a decorator-hostile decorator, :data:`None`.
            * If this attribute is a submodule, type, or instance transitively
              defining one or more decorator-hostile decorators, the attribute
              name subtrie to map this attribute name to.
        '''
        assert isinstance(attr_name, str), f'{repr(attr_name)} not string.'
        assert isinstance(
            attr_name_subtrie, NoneTypeOr[BeartypeDecorPlaceTrieABC]), (
            f'{repr(attr_name_subtrie)} neither "None" nor frozen dictionary.')

        # Render this scope's beforelist safe for modification if this
        # beforelist is *NOT* yet safely modifiable.
        self._scope.permute_beforelist_if_needed()

        # Fully-qualified name of the top-level root package or module
        # transitively containing the attribute with the passed name (e.g.,
        # "some_package" when "attr_name" is
        # "some_package.some_module.some_submodule").
        #
        # This root package name is *MUCH* more relevant than this attribute
        # name, which typically signifies the name of an external attribute
        # imported into the currently visited module. Trivial imports of the
        # form "import {package_name}" are similar to complex imports of the
        # form "import {package_name}.{submodule_name}.{attr_name}". In both
        # cases, Python only imports the top-level root package whose name is
        # "{package_name}" into the current lexical scope. For static analysis
        # purposes, trailing "."-prefixed substrings like
        # ".{submodule_name}.{attr_name}" are almost entirely irrelevant.
        #
        # Note this has been profiled to be the fastest one-liner parsing the
        # first "."-suffixed substring from a "."-delimited string.
        attr_basename = (
            attr_name.partition('.')[0] if '.' in attr_name else attr_name)
        # print(f'Mapping imported decorator-hostile decorator basename "{attr_basename}"...')
        # print(f'...to subtrie "{attr_name_subtrie}".')

        # Map the fully-qualified name of the top-level root package or module
        # transitively containing the attribute with this name to this subtrie.
        self._scope.beforelist.scoped_attr_basename_trie[attr_basename] = (  # type: ignore[index]
            attr_name_subtrie)

    # ....................{ PRIVATE ~ decorators           }....................
    #FIXME: Revise docstring, please. This method now employs a highly
    #non-trivial algorithm to decide the correct @beartype decorator position in
    #a chain of one or more existing non-@beartype decorators.
    #FIXME: Unit test us up, please.
    def _decorate_node_beartype(
        self, node: NodeDecoratable, conf: BeartypeConf) -> None:
        '''
        Add a new **child beartype decoration node** (i.e., abstract syntax tree
        (AST) node applying the :func:`beartype.beartype` decorator configured
        by the passed beartype configuration) to the passed **parent decoratable
        node** (i.e., AST node encapsulating the definition of a pure-Python
        object supporting decoration by one or more ``"@"``-prefixed
        decorations, including both pure-Python classes *and* callables).

        Note that this function **prepends** rather than appends this child
        decoration node to the beginning of the list of all child decoration
        nodes for this parent decoratable node. Since that list is "stored
        outermost first (i.e. the first in the list will be applied last)",
        prepending guarantees that the beartype decorator will be applied last
        (i.e., *after* all other decorators). This ensures that explicitly
        configured beartype decorations applied to this decoratable by the end
        user (e.g., ``@beartype(conf=BeartypeConf(...))``) assume precedence
        over implicitly configured beartype decorations applied by this
        function.

        Parameters
        ----------
        node : AST
            **Decoratable node** (i.e., parent class or callable node) to add a
            new child beartype decoration node to.
        conf : BeartypeConf
            **Beartype configuration** (i.e., dataclass configuring the
            :mod:`beartype.beartype` decorator for some decoratable object(s)
            decorated by a parent node passing this dataclass to that
            decorator).
        '''
        assert isinstance(node, AST), f'{repr(node)} not AST node.'
        assert isinstance(conf, BeartypeConf), (
            f'{repr(conf)} not configuration.')

        # ....................{ NODES                      }....................
        # Child decoration node decorating this parent type or callable node by
        # the @beartype decorator.
        node_beartype_decorator: expr = Name(
            id=BEARTYPE_DECORATOR_FUNC_NAME, ctx=NODE_CONTEXT_LOAD)

        # Copy all source code metadata from this parent type or callable node
        # onto this child decorator node.
        copy_node_metadata(node_src=node, node_trg=node_beartype_decorator)

        # ....................{ NODES ~ conf               }....................
        #FIXME: Isn't this pretty much *ALWAYS* the case? "beartype.claw" import
        #hooks apply a non-default beartype configuration by default, sadly.
        # If the current beartype configuration is *NOT* the default beartype
        # configuration, this configuration is a user-defined beartype
        # configuration which *MUST* be passed to a call to the @beartype
        # decorator. Merely referencing this decorator does *NOT* suffice. In
        # this case...
        if conf != BEARTYPE_CONF_DEFAULT:
            # Replace the reference to this decorator defined above with a call
            # to this decorator passed this configuration.
            node_beartype_decorator = Call(
                func=node_beartype_decorator,
                args=[],
                # Node encapsulating the passing of this configuration as the
                # "conf" keyword argument to this call.
                keywords=[self._make_node_keyword_conf(node_sibling=node)],  # type: ignore[attr-defined]
            )

            # Copy all source code metadata from this parent callable node onto
            # this child call node.
            copy_node_metadata(node_src=node, node_trg=node_beartype_decorator)
        # Else, this configuration is simply the default beartype configuration.
        # In this case, avoid passing that configuration to the @beartype
        # decorator for both efficiency and simplicity.

        # ....................{ POSITION                   }....................
        # Decorator position (i.e., location to which the @beartype decorator
        # will be implicitly injected into the existing chain of zero or more
        # decorators already decorating this parent decoratable node), defined
        # as either...
        decoration_position = (
            # If this parent decoratable node is a parent class node, the
            # class-specific decorator position.
            conf.claw_decor_place_type
            if isinstance(node, ClassDef) else
            # Else, this parent decoratable node is *NOT* a parent class node.
            # By process of elimination, this parent decoratable node *MUST* be
            # a parent callable node. In this case, fallback to the
            # callable-specific decorator position.
            conf.claw_decor_place_func
        )
        # print(f'Injecting @beartype decorator onto decoratable {unparse(node)} decorator chain...')

        # If injecting the @beartype decorator contextually last, insert this
        # child decoration node immediately *BEFORE* the first existing child
        # decoration node encapsulating a decorator-hostile decorator externally
        # configured by the beforelist and previously detected by the
        # visit_Import*() family of methods defined above.
        #
        # Note that:
        # * This is the default position and thus detected first.
        # * The "node.decorator_list" list of nodes describing the chain of all
        #   existing decorators decorating this type or callable is "stored
        #   outermost first (i.e. the first in the list will be applied last)."
        if decoration_position is (
            BeartypeDecorPlace.LAST_BEFORE_DECOR_HOSTILE):
            self._decorate_node_beartype_last_before_decor_hostile(
                node=node, conf=conf, node_beartype_decorator=node_beartype_decorator)
        # If injecting the @beartype decorator unconditionally last, prepend
        # this child decoration node to the beginning of the list of all child
        # decoration nodes for this parent decoratable node. Prepending
        # guarantees that our decorator will be applied last (i.e., *AFTER* all
        # subsequent decorators).
        elif decoration_position is BeartypeDecorPlace.LAST:
            node.decorator_list.insert(0, node_beartype_decorator)
        # If injecting the @beartype decorator unconditionally first, append
        # this child decoration node to the end of the list of all child
        # decoration nodes for this parent decoratable node.
        elif decoration_position is BeartypeDecorPlace.FIRST:
            node.decorator_list.append(node_beartype_decorator)
        # Else, an unrecognized decorator position was configured. In this case,
        # raise an exception. Note that this should *NEVER* occur.
        else:  # pragma: no cover
            raise BeartypeClawImportConfException(
                f'Beartype configuration {repr(conf)} '
                f'decorator position {repr(decoration_position)} unsupported.'
            )


    def _decorate_node_beartype_last_before_decor_hostile(
        self,
        node: NodeDecoratable,
        conf: BeartypeConf,
        node_beartype_decorator: expr,
    ) -> None:
        '''
        Add a new child :func:`beartype.beartype` decoration node to the passed
        parent decoratable node subject to
        :attr:`BeartypeDecorPlace.LAST_BEFORE_DECOR_HOSTILE` positioning.

        This method contextually injects the :func:`beartype.beartype` decorator
        as high (i.e., late) in the chain of decorators decorating the type
        or callable encapsulated by this parent decoratable node as feasible
        while still preserving compatibility with decorator-hostile decorators.
        To do so, this method injects this decorator immediately *before* the
        lowest (i.e., earliest) decorator-hostile decorator decorating this type
        or callable as externally configured by the beforelist and previously
        detected by the ``visit_Import*()`` family of methods defined above.

        Parameters
        ----------
        node : AST
            **Decoratable node** (i.e., parent type or callable node) to add a
            new child beartype decoration node to.
        conf : BeartypeConf
            **Beartype configuration** (i.e., dataclass configuring the
            :mod:`beartype.beartype` decorator for some decoratable object(s)
            decorated by a parent node passing this dataclass to that
            decorator).
        node_beartype_decorator : expr
            Child decoration node decorating this parent type or callable node
            by the :func:`beartype.beartype` decorator.
        '''
        assert isinstance(node_beartype_decorator, expr), (
            f'{repr(node_beartype_decorator)} not AST expression node.')

        # ..................{ LOCALS                         }..................
        # If either...
        #
        # Note that this is the common case. Since undecorated types and
        # callables *AND* decorator-hostile decorators are all the exception
        # rather than the rule, optimizing for this common case is both
        # desirable and useful.
        if (
            # The currently visited type or callable being decorated by the
            # @beartype decorator is currently undecorated *OR*...
            not node.decorator_list or
            # This type or callable is decorated.
            #
            # The currently visited module imports neither decorator-hostile
            # decorators *NOR* modules transitively defining such decorators...
            not self._scope.beforelist.scoped_attr_basename_trie
        ):
        # Then this type or callable is decorated by *NO* decorator-hostile
        # decorators. In this case, the @beartype decorator may be safely
        # injected as the last decorator in the chain of decorators
        # decorating this type or callable.
            # print(f'Trivially detected decoratable {unparse(node)} to be decorated by no hostile decorators!')

            # Reduce to the trivial implementation of the
            # "BeartypeDecorPlace.LAST" position in the parent
            # _decorate_node_beartype() method.
            node.decorator_list.insert(0, node_beartype_decorator)

            # Halt further processing.
            return
        # Else, this type or callable is decorated *AND* the currently visited
        # module imports either decorator-hostile decorators *OR* modules
        # transitively defining such decorators...

        # 0-based index of the current decorator visited by iteration below in
        # the chain of all decorators decorating this type or callable.
        #
        # Note that this index also efficiently doubles as the position in this
        # decorator chain that the @beartype decorator should be injected into.
        # This index is thus one larger than that of the last decorator-hostile
        # decorator in this decorator chain.
        node_decor_index_curr = 0

        # 0-based index of the last decorator visited by iteration below.
        #
        # Note that this index is guaranteed to be non-negative (i.e., >= 0), as
        # the chain of all decorators decorating this type or callable is
        # guaranteed to be non-empty (by the above validation).
        node_decor_index_last = len(node.decorator_list) - 1

        # ..................{ SEARCH                         }..................
        # For the 0-based index of each child node encapsulating an existing
        # decorator in the chain of all decorators decorating this type or
        # callable *AND* that child node (in descending order of the last to
        # first such decorator)...
        #
        # With respect to AST transformations, there exist four kinds of
        # decorations (in order of increasing complexity). This method *MUST*
        # transparently support each of these kinds:
        # * An unqualified attribute decoration of the form
        #   "@{decorator_name}", encapsulated by a simple "ast.Name" node.
        # * A qualified attribute decoration of the form
        #   "@{module_or_type_name}.{decorator_name}", encapsulated by a less
        #   simple "ast.Attribute" node.
        # * A callable call decoration encapsulating either:
        #   * An unqualified attribute decoration of the form
        #     "@{decorator_name}(...)", encapsulated by a complex
        #     "ast.Call(func=Name(...), ...)" node.
        #   * A qualified attribute decoration of the form
        #     "@{module_or_type_name}.{decorator_name}(...)", encapsulated by an
        #     even more complex "ast.Call(func=Attribute(...), ...)" node.
        #
        # Consider this minimal-length example:
        #
        #     @beartype
        #     @beartype.beartype
        #     @beartype(conf=BeartypeConf(is_debug=True))
        #     @beartype.beartype(conf=BeartypeConf(is_debug=True))
        #     def muh_fun(): pass
        #
        # ...which is encapsulated by this AST:
        #     FunctionDef(
        #       name='muh_fun',
        #       args=arguments(),
        #       body=[
        #           Pass()],
        #       decorator_list=[
        #           Name(id='beartype', ctx=Load()),
        #           Attribute(
        #               value=Name(id='beartype', ctx=Load()),
        #               attr='beartype',
        #               ctx=Load()),
        #           Call(
        #               func=Name(id='beartype', ctx=Load()),
        #               keywords=[
        #                   keyword(
        #                       arg='conf',
        #                       value=Call(
        #                           func=Name(id='BeartypeConf', ctx=Load()),
        #                           keywords=[
        #                               keyword(
        #                                   arg='is_debug',
        #                                   value=Constant(value=True))]))]),
        #           Call(
        #               func=Attribute(
        #                   value=Name(id='beartype', ctx=Load()),
        #                   attr='beartype',
        #                   ctx=Load()),
        #               keywords=[
        #                   keyword(
        #                       arg='conf',
        #                       value=Call(
        #                           func=Name(id='BeartypeConf', ctx=Load()),
        #                           keywords=[
        #                               keyword(
        #                                   arg='is_debug',
        #                                   value=Constant(value=True))]))])])])

        # While the 0-based index of the current decorator is less than or equal
        # to that of the last decorator in this decorator chain...
        while node_decor_index_curr <= node_decor_index_last:
            # Current decorator.
            node_decor = node.decorator_list[node_decor_index_curr]

            # If this decoration is encapsulated by a complex "Call" node,
            # this is a callable call decoration encapsulating either:
            # * An unqualified attribute decoration of the form
            #   "@{decorator_name}(...)".
            # * A qualified attribute decoration of the form
            #   "@{decorator_name}(...)".
            #
            # In either case, reduce this callable call node to the child "Name"
            # or "Attribute" node it encapsulates.
            if isinstance(node_decor, Call):
                node_decor = node_decor.func
            # Else, this decoration is *NOT* encapsulated by a "Call" node.
            # print(f'Detecting decorator @{unparse(node_decor)} hostility...')

            # Either:
            # * If this decorator name refers to a previously imported
            #   decorator-hostile decorator, "True".
            # * If this decorator name refers to a previously imported
            #   submodule, type, or instance transitively defining one or more
            #   decorator-hostile decorators, the imported decorator-hostile
            #   attribute name subtrie describing the contents of that
            #   submodule, type, or instance.
            # * Else (i.e., if this decorator name refers to neither a
            #   previously imported decorator-hostile decorator *nor* a
            #   submodule, type, or instance transitively defining these
            #   decorators), "False".
            is_decor_hostile = self._is_node_scoped_attr_name(node_decor)
            # print(f'Decorator @{unparse(node_decor)} hostility: {is_decor_hostile}')

            # If this decorator name refers to a previously imported submodule,
            # type, or instance transitively defining one or more
            # decorator-hostile decorators, this decorator is *NOT* actually a
            # decorator according to the beforelist. But this decorator is a
            # decorator, clearly! Ergo, the user improperly configured a
            # "beartype.claw" import hook with an erroneous beforelist
            # specifying this imported decorator-hostile decorator to actually
            # be a non-decorator submodule, type, or instance. An object can be
            # a decorator *OR* non-decorator -- but not both. In this case...
            if isinstance(is_decor_hostile, BeartypeDecorPlaceTrieABC):
                # Possibly fully-qualified name of this decorator in this
                # lexical scope, defined by "unparsing" this child node. The
                # standard ast.unparse() function "unparses" (i.e., obtains the
                # machine-readable representations of) arbitrary nodes.
                #
                # Note that the parent object of this attribute is described by
                # the external node "node_target.value", encapsulating an
                # arbitrarily complex Python expression. "Unparsing" this
                # expression manually is *ABSOLUTELY* infeasible.
                decor_name = unparse(node_decor)

                # Raise an explanatory exception.
                raise BeartypeClawAstImportException(
                    f'Beartype configuration {repr(self._conf)} '  # type: ignore[attr-defined]
                    f'decorator position beforelist '
                    f'{repr(self._scope.beforelist.schema_attr_basename_trie)} '
                    f'nested frozen dictionary of '
                    f'decorator-hostile decorator names '
                    f'{repr(is_decor_hostile)} erroneously implies '
                    f'object "{node.name}" decorator @{decor_name} '
                    f'to not be a decorator (but instead be a '
                    f'subpackage, submodule, '
                    f'non-decorator type, or '
                    f'non-decorator callable).'
                )
            # Else, this decorator name does *NOT* refers to a previously
            # imported submodule, type, or instance transitively defining one or
            # more decorator-hostile decorators.
            #
            # If this decorator is *NOT* decorator-hostile, this decorator is
            # presumably compatible with the @beartype decorator. In this case,
            # immediately halt this iteration.
            elif is_decor_hostile is False:
                # print(f'Decorator @{unparse(node_decor)} is friendly!')
                break
            # Else, this decorator is decorator-hostile. In this case, continue
            # searching for the first non-hostile decorator.
            # print(f'Decorator @{unparse(node_decor)} is hostile!')

            # Validate sanity.
            assert is_decor_hostile is True

            # Increment the 0-based index of the next decorator to be visited.
            node_decor_index_curr += 1

        # If the 0-based index in this decorator chain that the @beartype
        # decorator should be injected into is a valid index into this list,
        # inject the @beartype decorator as the last decorator *BEFORE* (i.e.,
        # below) the last decorator-hostile decorator in the chain of existing
        # decorators decorating this type or callable.
        if node_decor_index_curr <= node_decor_index_last:
            # print(f'Injecting @beartype into decorator index {node_decor_index_curr}!')
            node.decorator_list.insert(
                node_decor_index_curr, node_beartype_decorator)
        # Else, this is *NOT* a valid index into this list. Presumably, this
        # index is one larger than the last valid index into this list. In this
        # case, this type or callable is decorated *ONLY* by decorator-hostile
        # decorators. Since type or callable is decorated by *NO* safe
        # decorators that the @beartype decorator can be injected before,
        # @beartype *MUST* instead be appended after the last decorator (which
        # is also a decorator-hostile decorator) decorating this type or
        # callable. Look. It's complicated.
        else:
            # Sanity-check the expected constraint.
            assert node_decor_index_curr == node_decor_index_last + 1, (
                f'Type or callable "{self._module_name}.{node.name}" '  # type: ignore[attr-defined]
                f'decorated by decorator-hostile decorators, but '
                f'@beartype decoration chain insertion index '
                f'{node_decor_index_curr} != {node_decor_index_last + 1}.'
            )
            # print('Appending @beartype after existing last decorator!')

            # Append @beartype *AFTER* this last decorator.
            node.decorator_list.append(node_beartype_decorator)
        # print(f'Decorator list after @beartype injection: {unparse(node.decorator_list)}')

    # ....................{ PRIVATE ~ finders              }....................
    def _is_node_scoped_attr_name(self, node: AST) -> Union[
        BeartypeDecorPlaceSubtrie, bool]:
        '''
        :data:`True` if the attribute name (presumably referenced from the
        current scope of the currently visited module) encapsulated by the
        passed node is that of a previously imported decorator-hostile
        decorator, an **imported decorator-hostile attribute name subtrie**
        (i.e., recursive tree structure whose nodes are the unqualified
        basenames of third-party attributes imported into a scope of the
        currently visited module such that these attributes are either
        themselves decorator-hostile decorators *or* submodules, types, or
        instances transitively defining decorator-hostile decorators) if that
        attribute name is that of a previously imported submodule, type, or
        instance transitively defining one or more decorator-hostile decorators,
        or :data:`False` otherwise (i.e., if this attribute name is *not* that
        of either a previously imported decorator-hostile decorator *or* a
        submodule, type, or instance transitively defining such decorators).

        This finder identifies whether the passed attribute name refers to a
        previously imported decorator-hostile decorator or not. Specifically,
        this finder (in order):

        #. Splits the possibly fully-qualified attribute name encapsulated by
           this attribute name node on ``"."`` delimiters into its constituent
           unqualified basenames.
        #. For each such unqualified basename:

           * If there exists a child subtrie of the current parent imported
             decorator-hostile attribute name (sub)trie (starting at the root
             decorator-hostile attribute name trie) whose associated key is
             this basename:

             * If that child subtrie is :data:`None` (signifying that child
               subtrie to be a terminal leaf node and thus a decorator-hostile
               decorator function or method), return :data:`True`.
             * Else, recurse into that subtrie.

           * Else, return either:

             * If this is the first unqualified basename and no such child
               subtrie exists, :data:`False`.
             * Else, that child subtrie.

        Parameters
        ----------
        node : AST
            Node possibly encapsulating the name of an attribute referenced from
            some scope of the currently visited module.

        Returns
        -------
        Optional[Union[BeartypeDecorPlaceSubtrie, Iota]]
            Either:

            * If this attribute name refers to a previously imported
              decorator-hostile decorator, :data:`True`.
            * If this attribute name refers to a previously imported submodule,
              type, or instance transitively defining one or more
              decorator-hostile decorators, the imported decorator-hostile
              attribute name subtrie describing the contents of that submodule,
              type, or instance.
            * Else (i.e., if this attribute name refers to neither a previously
              imported decorator-hostile decorator *nor* a submodule, type, or
              instance transitively defining these decorators),
              :data:`False`.
        '''

        # Current child (sub)trie of the parent scoped attribute name (sub)trie
        # being recursed into by iteration below, initialized to the scoped
        # attribute name trie unique to the current scope of the currently
        # visited module.
        attr_name_subtrie = self._scope.beforelist.scoped_attr_basename_trie
        # print(f'[_is_node_scoped_attr_name] Root trie: {attr_name_subtrie}')

        # If *NO* decorator-hostile attributes have been imported into the
        # current scope, silently reduce to a noop by returning false.
        if not attr_name_subtrie:
            # print(f'Trivially ignoring friendly attribute "{unparse(node)}" (due to empty trie)!')
            return False
        # Else, one or more decorator-hostile attributes have been imported into
        # the current scope.

        # Replace the contents of the existing "_attr_basenames" list with the
        # one or more unqualified basenames comprising the possibly
        # fully-qualified name of this attribute.
        self._attr_basenames = get_node_attr_basenames(
            node=node, attr_basenames=self._attr_basenames)
        # print(f'Split attribute name "{unparse(node)}" into basenames {repr(self._attr_basenames)}.')

        # If this list is empty, the prior getter failed to reconstruct this
        # list from this node (e.g., due to this node being of an unknown type
        # currently unsupported by that getter). This implies it is now unknown
        # whether the desired attribute is has been imported from a problematic
        # third-party package into the currently visited module or not.
        #
        # Note that this should *NEVER* happen. The prior getter *SHOULD*
        # support all relevant types of nodes. However, the Python language and
        # hence AST grammar describing that language is constantly evolving.
        # Since this just happened, it is likely that a future iteration of the
        # Python language has now evolved in an unanticipated (yet, ultimately
        # valid) way. To preserve forward compatibility in @beartype with future
        # Python versions, intentionally ignore this decorator.
        #
        # Sometimes, doing nothing at all is the best thing you can do.
        if not self._attr_basenames:
            # print(f'Trivially ignoring friendly attribute "{unparse(node)}" (due to unparseable basenames)!')
            # print(f'AST node structure:\n{get_node_repr_indented(node)}')
            return False
        # Else, this list is non-empty, implying the prior getter succeeded in
        # reconstructing this list from this node.

        # For the unqualified basename of either each parent submodule
        # transitively defining this attribute itself or this attribute...
        for attr_basename in self._attr_basenames:
            # print(f'Mapping attribute basename "{attr_basename}"...')

            # Child subtrie of this parent (sub)trie matching this basename if
            # the currently visited module previously imported either a
            # third-party submodule, type, or instance transitively defining one
            # or more decorator-hostile decorators or such a decorator with the
            # same basename *OR* the sentinel.
            attr_name_subtrie = attr_name_subtrie.get(attr_basename, SENTINEL)  # type: ignore[arg-type, assignment]
            # print(f'...to subtrie "{attr_name_subtrie}".')

            # If the parent (sub)trie contained *NO* basename matching that of a
            # problematic third-party object imported into the currently visited
            # module, this attribute is *NOT* a decorator-hostile decorator.
            # This attribute is thus presumably friendly to the @beartype
            # decorator. In this case, immediately return false.
            #
            # Note that:
            # * *ALL* decorators syntactically following (i.e., applied
            #   semantically earlier to the currently decorated type or
            #   callable) the top-most decorator-friendly decorator in a chain
            #   of decorators *MUST* themselves be decorator-friendly. Why?
            #   Because, if any following decorator was decorator-hostile, then
            #   by definition that decorator-hostile decorator could *NOT* have
            #   been preceded by any other decorators. That's what
            #   decorator-hostile means. But that decorator *WAS* preceded by
            #   this decorator. Proof by contradiction yields this conclusion.
            if attr_name_subtrie is SENTINEL:
                # print(f'Detected attribute name "{unparse(node)}" as unmatched decorator-friendly!')
                return False
            # Else, the parent (sub)trie contained a basename matching that of a
            # problematic third-party object imported into the currently visited
            # module, implying this attribute *COULD* still be hostile.
            #
            # If this child subtrie is "None", the key-value pair of that parent
            # (sub)trie associated with this "None" value is a terminal leaf
            # node, implying this attribute to be a decorator-hostile decorator.
            # In this case, immediately return true.
            elif attr_name_subtrie is None:
                # print(f'Detected attribute name "{unparse(node)}" as decorator-hostile decorator!')
                return True
            # Else, this child subtrie is *NOT* "None". By elimination, this
            # child subtrie *MUST* actually be yet another recursively nested
            # subtrie. In this case, continue matching basenames between this
            # attribute and decorator-hostile decorators imported into the
            # currently visited module.
        # Else, the list of all unqualified basenames comprising the possibly
        # fully-qualified name of this attribute was exhausted:
        # * *AFTER* recursively visiting at least one imported attribute name
        #   (sub)trie with corresponding basenames. Why? Because if at least one
        #   (sub)trie had *NOT* been visited above, this method would have
        #   immediately returned false. But this method has yet to return
        #   anything! Proof by contradiction yields this conclusion.
        # * *BEFORE* recursively visiting all imported attribute name (sub)tries
        #   with associated basenames. Why? Because if all such (sub)tries had
        #   already been visited above, this method would have immediately
        #   returned true. But this method has yet to return anything! QED.
        # print(f'Detected attribute name "{unparse(node)}" as decorator-hostile subtrie {attr_name_subtrie}!')

        # Assert sanity. You never know, bear friends. And neither do we.
        assert isinstance(attr_name_subtrie, BeartypeDecorPlaceTrieABC)

        # Return the most recently visited imported attribute name (sub)trie.
        return attr_name_subtrie



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) fixtures** (i.e.,
:mod:`pytest`-specific context managers passed as parameters to unit tests
exercising the :mod:`beartype.door.TypeHint` superclass).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from pytest import fixture

# ....................{ FIXTURES ~ equality                }....................
@fixture(scope='session')
def door_cases_equality() -> 'Iterable[Tuple[object, object, bool]]':
    '''
    Session-scoped fixture returning an iterable of **hint equality cases**
    (i.e., 3-tuples ``(hint_a, hint_b, is_equal)`` describing the equality
    relations between two PEP-compliant type hints), efficiently cached across
    all tests requiring this fixture.

    This iterable is intentionally defined by the return of this fixture rather
    than as a global constant of this submodule. Why? Because the former safely
    defers all heavyweight imports required to define this iterable to the call
    of the first unit test requiring this fixture, whereas the latter unsafely
    performs those imports at pytest test collection time.

    Returns
    -------
    Iterable[Tuple[object, object, bool]]
        Iterable of one or more 3-tuples ``(hint_a, hint_b, is_equal)``,
        where:

        * ``hint_a`` is the PEP-compliant type hint to be passed as the first
          parameter to the :meth:`beartype.door.TypeHint.__equals__` tester.
        * ``hint_b`` is the PEP-compliant type hint to be passed as the second
          parameter to the :meth:`beartype.door.TypeHint.__equals__` tester.
        * ``is_equal`` is ``True`` only if these hints are equal according to
          that tester.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype.typing import Annotated
    from collections.abc import (
        Awaitable as AwaitableABC,
        Sequence as SequenceABC,
    )
    from numbers import Number

    # Intentionally import from "typing" rather than "beartype.typing" to
    # guarantee PEP 484-compliant type hints.
    from typing import (
        Any,
        List,
        Tuple,
        Union,
    )

    # ..................{ LISTS                              }..................
    HINT_EQUALITY_CASES = [
        # ..................{ PEP 484 ~ argless : bare       }..................
        # PEP 484-compliant unsubscripted type hints, which are necessarily
        # equal to themselves.
        (tuple, Tuple, True),
        (list, list, True),
        (list, List, True),

        # ..................{ PEP 484 ~ arg : sequence       }..................
        # PEP 484-compliant sequence type hints.
        (list, List[Any], True),
        (tuple, Tuple[Any, ...], True),

        # ..................{ PEP 484 ~ arg : union          }..................
        # PEP 484-compliant union type hints.
        (Union[int, str], Union[str, list], False),
        (Union[Number, int], Union[Number, float], True),

        # Test that union equality ignores order.
        (Union[int, str], Union[str, int], True),

        # Test that union equality compares child type hints collectively rather
        # than individually.
        #
        # Note that this pair of cases tests numerous edge cases, including:
        # * Equality comparison of non-unions against unions. Although
        #   "Union[int]" superficially appears to be a union, Python reduces
        #   "Union[int]" to simply "int" at runtime.
        (Union[bool, int], Union[int], True),
        (Union[int], Union[bool, int], True),

        # ..................{ PEP 585                        }..................
        # PEP 585-compliant type hints.
        (tuple[str, ...], Tuple[str, ...], True),
        (list[str], List[str], True),
        (AwaitableABC[SequenceABC[int]], AwaitableABC[SequenceABC[int]], True),

        # ..................{ PEP 593                        }..................
        # PEP 593-compliant type hints.
        (Annotated[int, "hi"], Annotated[int, "hi"], True),
        (Annotated[int, "hi"], Annotated[int, "low"], False),
        (Annotated[int, "hi"], Annotated[int, "low"], False),
    ]

    # Return this mutable list coerced into an immutable tuple for safety.
    return tuple(HINT_EQUALITY_CASES)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`593`-compliant **type metahint reducers** (i.e., low-level
low-level callables converting higher-level type hints created by subscripting
the :obj:`typing.Annotated` type hint factory to lower-level type hints more
readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatypingport import Hint
from beartype._util.hint.pep.proposal.pep593 import (
    get_hint_pep593_metahint,
    is_hint_pep593_beartype,
)

# ....................{ REDUCERS                           }....................
def reduce_hint_pep593(hint: Hint, exception_prefix: str) -> Hint:
    '''
    Reduce the passed :pep:`593`-compliant **type metahint** (i.e., subscription
    of the :obj:`typing.Annotated` hint factory) to a lower-level hint if this
    metahint contains *no* **beartype validators** (i.e., subscriptions of
    :mod:`beartype.vale` factories).

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to a
    one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    Hint
        Lower-level type hint currently supported by :mod:`beartype`.
    '''
    # print(f'Reducing non-beartype PEP 593 type hint {repr(hint)}...')

    # Return either...
    return (
        # If this metahint is beartype-specific, preserve this hint as is for
        # subsequent handling elsewhere;
        hint
        if is_hint_pep593_beartype(hint) else
        # Else, this metahint is beartype-agnostic and thus irrelevant to us. In
        # this case, ignore all annotations on this hint by reducing this hint
        # to the lower-level hint it annotates.
        get_hint_pep593_metahint(hint=hint, exception_prefix=exception_prefix)
    )
